from typing import List
import pandas as pd
import numpy as np
from .engineobj import SqPandasEngine
from suzieq.sqobjects import get_sqobject
from suzieq.utils import convert_macaddr_format_to_colon


class NetworkObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'network'

    def get(self, **kwargs):
        """Get the information requested"""

        view = kwargs.get('view', self.iobj.view)
        columns = kwargs.pop('columns', ['default'])
        addnl_fields = kwargs.pop('addnl_fields', [])
        user_query = kwargs.pop('query_str', '')
        os = kwargs.pop('os', [])
        model = kwargs.pop('model', [])
        vendor = kwargs.pop('vendor', [])
        os_version = kwargs.pop('version', [])
        namespace = kwargs.pop('namespace', [])

        drop_cols = []

        if os or model or vendor or os_version:
            df = get_sqobject('device')(context=self.ctxt).get(
                columns=['namespace', 'hostname', 'os', 'model', 'vendor',
                         'version'],
                os=os,
                model=model,
                vendor=vendor,
                version=os_version,
                namespace=namespace,
                **kwargs)
        else:
            df = get_sqobject('device')(context=self.ctxt).get(
                columns=['namespace', 'hostname'], namespace=namespace,
                **kwargs)

        if df.empty:
            return pd.DataFrame()

        namespace = df.namespace.unique().tolist()

        # Get list of namespaces we're polling
        pollerdf = get_sqobject('sqPoller')(context=self.ctxt) \
            .get(columns=['namespace', 'hostname', 'service', 'status', 'timestamp'],
                 namespace=namespace,
                 **kwargs)
        if pollerdf.empty:
            return pd.DataFrame()

        nsgrp = pollerdf.groupby(by=['namespace'])
        pollerns = sorted(pollerdf.namespace.unique().tolist())
        newdf = pd.DataFrame({
            'namespace': pollerns,
            'deviceCnt': nsgrp['hostname'].nunique().tolist(),
            'serviceCnt': nsgrp['service'].nunique().tolist()
        })

        errsvc_df = pollerdf.query('status != 0 and status != 200') \
                            .groupby(by=['namespace'])['service'] \
                            .nunique().reset_index()
        newdf = newdf.merge(errsvc_df, on=['namespace'], how='left',
                            suffixes=['', '_y']) \
            .rename({'service': 'errSvcCnt'}, axis=1) \
            .fillna({'errSvcCnt': 0})
        newdf['errSvcCnt'] = newdf['errSvcCnt'].astype(int)

        # What protocols exist
        for table, fld in [('ospf', 'hasOspf'), ('bgp', 'hasBgp'),
                           ('evpnVni', 'hasVxlan'), ('mlag', 'hasMlag')]:
            df = get_sqobject(table)(context=self.ctxt) \
                .get(namespace=pollerns, columns=['namespace'])
            if df.empty:
                newdf[fld] = False
                continue

            gotns = df.namespace.unique().tolist()
            newdf[fld] = newdf.apply(
                lambda x, y: True if x.namespace in y else False,
                axis=1, args=(gotns,))

        newdf = self._handle_user_query_str(newdf, user_query)

        # Look for the rest of info only in selected namepaces
        newdf['lastUpdate'] = nsgrp['timestamp'].max() \
            .reset_index()['timestamp']

        return newdf.drop(columns=drop_cols)

    def find(self, **kwargs):
        '''Find the information requsted:

        address: given a MAC or IP address, find its owner OR the first hop
                 device its connected to
        asn: given an ASN, find the list of BGP sessions associated with this ASN
        '''

        addrlist = kwargs.pop('address', [])
        asn = kwargs.pop('asn', '')
        columns = kwargs.pop('columns', ['default'])
        resolve_bond = kwargs.pop('resolve_bond', False)

        dflist = []
        if isinstance(addrlist, str):
            addrlist = [addrlist]
        for addr in addrlist:
            df = self._find_address(addr, resolve_bond, **kwargs)
            if not df.empty:
                dflist.append(df)
        if dflist:
            df = pd.concat(dflist)
        else:
            df = self._find_asn(asn, **kwargs)

        return df

    def summarize(self, **kwargs):
        '''Summarize for network

        Summarize for network is a bit different from the rest of the
        summaries because its not grouped by namespace.
        '''

        df = self.get(**kwargs)

        if df.empty:
            return df

        self.ns = pd.DataFrame({
            'namespacesCnt': [df.namespace.count()],
            'servicePerNsStat': [[df.serviceCnt.min(), df.serviceCnt.max(),
                                  df.serviceCnt.median()]],
            'nsWithMlagCnt': [df.loc[df.hasMlag].shape[0]],
            'nsWithBgpCnt': [df.loc[df.hasBgp].shape[0]],
            'nsWithOspfCnt': [df.loc[df.hasOspf].shape[0]],
            'nsWithVxlanCnt': [df.loc[df.hasVxlan].shape[0]],
            'nsWithErrsvcCnt': [df.loc[df.errSvcCnt != 0].shape[0]],
        })

        # At this point we have a single row with index 0, so rename
        return self.ns.T.rename(columns={0: 'summary'})

    def _find_asn(self, asn: List[str], **kwargs) -> pd.DataFrame:
        """Find the hosts with the ASN listed

        Args:
            asn (List[str]): List of ASNs, space separated

        Returns:
            pd.DataFrame: The dataframe with the info. The columns are-
                          namespace, hostname, asn
        """

        if not asn:
            return pd.DataFrame()

        _ = kwargs.pop('vlan', [])
        try:
            asn = [int(x) for x in asn]
        except ValueError:
            return pd.DataFrame({'error': ['ASNs must be an integers']})

        return get_sqobject('bgp')(context=self.ctxt).get(asn=asn, columns=['default'],
                                                          **kwargs)

    def _find_address(self, addr: str, resolve_bond: bool,
                      **kwargs) -> pd.DataFrame:
        """Find the origin or the first hop network device that owns this addr

        Args:
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with the relevant information
        """

        cols = ['namespace', 'hostname', 'ifname',
                'vrf', 'ipAddress', 'vlan', 'macaddr', 'how', 'timestamp']

        vlan = kwargs.pop('vlan', '')
        vrf = kwargs.pop('vrf', '')

        # Convert Cisco-style MAC address to standard MAC addr format,
        # and lowercase all letters of the alphabet
        if '.' in addr:
            addr = convert_macaddr_format_to_colon(addr)
        addr = addr.lower()

        addr_df = get_sqobject('address')(context=self.ctxt).get(
            address=addr.split(),
            columns=['namespace', 'hostname', 'ifname', 'vrf', 'ipAddressList',
                     'ip6AddressList', 'macaddr', 'vlan', 'timestamp'],
            **kwargs)

        if addr_df.empty:
            addr_df = self._find_addr_arp(addr, **kwargs)
            addr_df['how'] = 'derived'
            do_continue = True
        else:
            do_continue = False
            addr_df['how'] = 'polled'
            addr_df = addr_df.explode('ipAddressList').fillna('')

        if not addr_df.empty:
            if do_continue:
                # Not polled entry, and so we need to follow L2 links
                addr_df = self._find_l2_attach(addr_df, **kwargs)
            if vlan:
                addr_df = addr_df.query(f'vlan == {vlan}')
            if vrf:
                addr_df = addr_df.query(f'vrf == "{vrf}"')
            if resolve_bond:
                addr_df = self._find_bond_slaves(addr_df
                                                 .reset_index(drop=True))
            return addr_df.rename(
                columns={'ipAddressList': 'ipAddress',
                         'ip6AddressList': 'ip6Address'}, errors='ignore')[cols] \
                .dropna()
        return addr_df

    def _find_bond_slaves(self, df: pd.DataFrame) -> pd.DataFrame:
        """This function splits out the bond ports if requested

        Assumes that the field ifname exists in the dataframe

        Args:
            df (pd.DataFrame): The dataframe with the field that needs
                               to be resolved

        Returns:
            pd.DataFrame: The dataframe with the bond name resolved if
                          possible, else unchanged
        """

        # get list  of namespaces, hostnames and ifnames to get info for
        ns = df.namespace.unique().tolist()
        hosts = df.hostname.unique().tolist()
        master = df.ifname.unique().tolist()

        ifdf = get_sqobject('interfaces')(context=self.ctxt) \
            .get(namespace=ns, hostname=hosts, master=master)

        if ifdf.empty:
            return df

        for i, row in enumerate(df.itertuples()):
            newiflist = ifdf.apply(
                lambda x, y: x.ifname
                if (x.master == row.ifname) and (x.namespace == row.namespace) and (x.hostname == row.hostname) else '',
                args=(row, ), axis=1).tolist()
            if newiflist:
                df.at[i, 'ifname'] = ', '.join(filter(None, newiflist))

        return df

    def _find_addr_arp(self, addr: str, **kwargs) -> pd.DataFrame:
        """Find the origin or the first hop network device that owns this addr

        Args:
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with the relevant information
        """

        cols = ['namespace', 'hostname', 'ifname',
                'vrf', 'ipAddress', 'macaddr', 'timestamp']

        if any(x in addr for x in ['::', '.']):
            arpdf = get_sqobject('arpnd')(context=self.ctxt).get(
                ipAddress=addr, **kwargs)
        else:
            arpdf = get_sqobject('arpnd')(context=self.ctxt).get(
                macaddr=addr, **kwargs)

        if not arpdf.empty:
            ifdf = get_sqobject('interfaces')(context=self.ctxt) \
                .get(namespace=arpdf.namespace.unique().tolist(),
                     hostname=arpdf.hostname.unique().tolist(),
                     ifname=arpdf.oif.unique().tolist())

            arpdf = arpdf.merge(ifdf[['namespace', 'hostname', 'ifname', 'vlan',
                                      'master']],
                                left_on=['namespace', 'hostname', 'oif'],
                                right_on=['namespace',
                                          'hostname', 'ifname'],
                                how='left', suffixes=['', '_y']) \
                .drop(columns=['oif']) \
                .rename(columns={'master': 'vrf'})

            macdf = get_sqobject('macs')(context=self.ctxt) \
                .get(namespace=arpdf.namespace.unique().tolist(),
                     hostname=arpdf.hostname.unique().tolist(),
                     vlan=arpdf.vlan.astype(str).unique().tolist(),
                     macaddr=arpdf.macaddr.unique().tolist(),
                     columns=['default'],
                     localOnly=True)
            if not macdf.empty:
                arpdf = arpdf.merge(macdf,
                                    on=['namespace', 'hostname',
                                        'vlan', 'macaddr'],
                                    how='left', suffixes=['', '_y']) \
                    .drop(columns=['remoteVtepIp', 'ifname',
                                   'timestamp_y'], errors='ignore') \
                    .rename(columns={'oif': 'ifname'})

            return arpdf['namespace hostname vrf ipAddress macaddr '
                         'vlan ifname timestamp'.split()]

        return arpdf

    def _find_l2_attach(self, addr_df: pd.DataFrame, **kwargs) -> pd.DataFrame:
        """Find the L2 attachment point for this address

        Args:
            addr_df (pd.DataFrame): Dataframe with the address information
            **kwargs: Additional arguments to pass to the sqobject

        Returns:
            pd.DataFrame: Dataframe with the L2 attachment point information
        """

        # We'll need to resolve bond interfaces first
        l2_addr_df = addr_df

        lldp_df = get_sqobject('lldp')(context=self.ctxt).get(
            namespace=l2_addr_df.namespace.unique().tolist(),
            columns=['namespace', 'hostname', 'ifname', 'peerHostname',
                     'peerIfname'])

        if lldp_df.empty:
            return addr_df

        do_continue = True
        while do_continue:
            do_continue = False
            # Do not attempt to replace tmp_df with l2_addr_df and get
            # rid of do_contiue. For some weird reason, with pandas ver
            # 1.2.5, the explode line below hangs if we just use l2_addr_df

            tmp_df = self._find_bond_slaves(l2_addr_df)
            tmp_df['ifname'] = tmp_df.ifname.str.split(', ')
            tmp_df = tmp_df.explode('ifname')

            new_l2_addr_df = tmp_df.merge(lldp_df,
                                          on=['namespace',
                                              'hostname', 'ifname'],
                                          suffixes=['', '_y'])
            if not new_l2_addr_df.empty:
                new_l2_addr_df = new_l2_addr_df.drop(
                    columns=['timestamp_y', 'hostname', 'ifname'],
                    errors='ignore') \
                    .rename(columns={'peerHostname': 'hostname',
                                     'peerIfname': 'ifname'})
                macdf = get_sqobject('macs')(context=self.ctxt) \
                    .get(namespace=new_l2_addr_df.namespace.unique().tolist(),
                         macaddr=new_l2_addr_df.macaddr.unique().tolist(),
                         vlan=new_l2_addr_df.vlan.astype(
                             str).unique().tolist(),
                         columns=['namespace', 'hostname', 'vlan', 'macaddr', 'oif'])
                if not macdf.empty:
                    new_l2_addr_df = new_l2_addr_df.merge(
                        macdf,
                        on=['namespace', 'hostname', 'vlan', 'macaddr'],
                        suffixes=['', '_y'])
                    if not new_l2_addr_df.empty:
                        new_l2_addr_df = new_l2_addr_df \
                            .drop(columns=['ifname', 'mackey', 'timestamp_y',
                                           'flags', 'remoteVtepIp'],
                                  errors='ignore') \
                            .rename(columns={'oif': 'ifname'})
                        l2_addr_df = new_l2_addr_df
                        do_continue = True
            else:
                l2_addr_df = tmp_df
                do_continue = False

            if not new_l2_addr_df.empty:
                l2_addr_df = new_l2_addr_df
            else:
                l2_addr_df = tmp_df

        return l2_addr_df.drop_duplicates(
            subset=['namespace', 'hostname', 'ifname', 'macaddr', 'vlan'])



#!/usr/bin/env python3

import sys
import os
import argparse
from time import sleep
import asyncio
import logging
from pathlib import Path
from collections import defaultdict
import getpass
from importlib.util import find_spec
import fcntl
import signal
import errno

import uvloop

from suzieq.poller.nodes import init_hosts, init_files
from suzieq.poller.services import init_services

from suzieq.poller.writer import init_output_workers, run_output_worker
from suzieq.utils import (load_sq_config, init_logger, ensure_single_instance,
                          get_sq_install_dir, get_log_params)


async def process_signal(signum, loop):
    tasks = [t for t in asyncio.all_tasks() if t is not
             asyncio.current_task()]

    for task in tasks:
        task.cancel()


def validate_parquet_args(cfg, output_args, logger):
    """Validate user arguments for parquet output"""

    if not cfg.get("data-directory", None):
        output_dir = "/tmp/suzieq/parquet-out/"
        logger.warning(
            "No output directory for parquet specified, using" "/tmp/suzieq/parquet-out"
        )
    else:
        output_dir = cfg["data-directory"]

    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    if not os.path.isdir(output_dir):
        logger.error(
            "Output directory {} is not a directory".format(output_dir))
        print("Output directory {} is not a directory".format(output_dir))
        sys.exit(1)

    logger.info("Parquet outputs will be under {}".format(output_dir))
    output_args.update({"output_dir": output_dir})

    return


async def start_and_monitor_coalescer(config_file: str, cfg: dict,
                                      logger: logging.Logger,
                                      coalescer_bin: str = None) -> None:
    '''Start and monitor the coalescer

    :param config_file: str, the path to suzieq config file, to be passed
    :param cfg: dict, the Suzieq config dictionary
    :param logger: logging.Logger, pointer to logger to use
    :param coalescer_bin: str, optional path to coalescer binary

    :return: nothing

    '''

    async def start_coalescer():
        sq_path = get_sq_install_dir()
        coalescer_bin = f'{sq_path}/utilities/sq_coalescer.py'
        if config_file:
            coalescer_args = f'-c {config_file}'
        else:
            coalescer_args = ''
        coalescer_args = f'{coalescer_bin} {coalescer_args}'.strip().split()

        try:
            process = await asyncio.create_subprocess_exec(
                *coalescer_args, stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE)
        except Exception as ex:
            logger.error(f'ABORTING. Unable to start coalescer: {ex}')
            process = None

        return process

    fd = 0
    process = None
    # Check to see file lock is possible
    while not fd:
        if not process:
            logger.warning('Starting Coalescer')
        elif process.returncode == errno.EBUSY:
            logger.warning('Trying to start coalescer')
        process = await start_coalescer()

        if not process:
            os.kill(os.getpid(), signal.SIGTERM)
            return

        # Initial sleep to ensure that the coalescer starts up
        await asyncio.sleep(10)
        coalesce_dir = cfg.get('coalescer', {})\
            .get('coalesce-directory',
                 f'{cfg.get("data-directory")}/coalesced')

        fd = ensure_single_instance(f'{coalesce_dir}/.sq-coalescer.pid',
                                    False)
        if fd > 0:
            # unlock and try to start process
            try:
                fcntl.flock(fd, fcntl.F_UNLCK)
                os.close(fd)
            except OSError:
                pass
            continue

        # Check if we have something from the stdout we need to log
        try:
            stdout, stderr = await process.communicate()
        except asyncio.CancelledError:
            if process:
                process.terminate()
                sleep(5)
                process.kill()
            return

        if process.returncode and (process.returncode != errno.EBUSY):
            logger.error(f'coalescer stdout: {stdout}, stderr: {stderr}')
        else:
            if process.returncode == errno.EBUSY:
                await asyncio.sleep(10*60)
            else:
                logger.info(
                    f'coalescer ended stdout: {stdout}, stderr: {stderr}')

        fd = 0


async def start_poller(userargs, cfg):

    logfile, loglevel, logsize, log_stdout = get_log_params(
        'poller', cfg, '/tmp/sq-poller.log')
    logger = init_logger('suzieq.poller', logfile,
                         loglevel, logsize, log_stdout)

    if userargs.devices_file and userargs.namespace:
        logger.error("Cannot specify both -D and -n options")
        sys.exit(1)

    if not os.path.exists(cfg["service-directory"]):
        logger.error(
            "Service directory {} is not a directory".format(
                cfg['service-directory'])
        )
        print("Service directory {} is not a directory".format(
            cfg['service-directory']))
        sys.exit(1)

    if userargs.jump_host and not userargs.jump_host.startswith('//'):
        logger.error("Jump host format is //<username>@<jumphost>:<port>")
        print("Jump host format is //<username>@<jumphost>:<port>")
        sys.exit(1)

    if not cfg.get("schema-directory", None):
        schema_dir = "{}/{}".format(userargs.service_dir, "schema")
    else:
        schema_dir = cfg["schema-directory"]

    if userargs.ssh_config_file:
        ssh_config_file = os.path.expanduser(userargs.ssh_config_file)
        if (os.stat(
                os.path.dirname(ssh_config_file)).st_mode | 0o40700 != 0o40700):
            logger.error('ssh directory has wrong permissions, must be 0700')
            print('ERROR: ssh directory has wrong permissions, must be 0700')
            sys.exit(1)

    output_args = {}

    if "parquet" in userargs.outputs:
        validate_parquet_args(cfg, output_args, logger)

    if userargs.run_once:
        userargs.outputs = ["gather"]
        output_args['output_dir'] = userargs.output_dir

    # Disable coalescer in specific, unusual cases
    # in case of input_dir, we also seem to leave a coalescer instance running
    if userargs.run_once or userargs.input_dir:
        userargs.no_coalescer = True

    outputs = init_output_workers(userargs.outputs, output_args)

    queue = asyncio.Queue()

    logger.info("Initializing hosts and services")
    tasks = []

    ignore_known_hosts = userargs.ignore_known_hosts or None

    if userargs.service_only:
        svclist = userargs.service_only.split()
    else:
        svcs = list(Path(cfg["service-directory"]).glob('*.yml'))
        svclist = [os.path.basename(x).split('.')[0] for x in svcs]

    if userargs.exclude_services:
        svclist = list(filter(lambda x: x not in userargs.exclude_services,
                              svclist))

    if not svclist:
        print(
            f"No correct services specified. Should have been one of {[svc.name for svc in svcs]}")
        sys.exit(1)

    connect_timeout = cfg.get('poller', {}).get('connect-timeout', 15)
    if userargs.input_dir:
        tasks.append(init_files(userargs.input_dir))
    else:
        tasks.append(init_hosts(inventory=userargs.devices_file,
                                ans_inventory=userargs.ansible_file,
                                namespace=userargs.namespace,
                                passphrase=userargs.passphrase,
                                jump_host=userargs.jump_host,
                                jump_host_key_file=userargs.jump_host_key_file,
                                password=userargs.ask_pass,
                                connect_timeout=connect_timeout,
                                ssh_config_file=userargs.ssh_config_file,
                                ignore_known_hosts=ignore_known_hosts))

    period = cfg.get('poller', {}).get('period', 15)

    tasks.append(init_services(cfg["service-directory"], schema_dir, queue,
                               svclist, period,
                               userargs.run_once or "forever"))

    nodes, svcs = await asyncio.gather(*tasks)

    if not nodes or not svcs:
        # Logging should've been done by init_nodes/services for details
        print('Termminating because no nodes or services found')
        sys.exit(0)

    node_callq = defaultdict(lambda: defaultdict(dict))

    node_callq.update({x: {'hostname': nodes[x].hostname,
                           'postq': nodes[x].post_commands}
                       for x in nodes})

    for svc in svcs:
        svc.set_nodes(node_callq)

    logger.setLevel(logging.INFO)
    logger.info("Suzieq Started")
    logger.setLevel(loglevel.upper())

    loop = asyncio.get_event_loop()
    for s in [signal.SIGTERM, signal.SIGINT]:
        loop.add_signal_handler(
            s, lambda s=s: asyncio.create_task(process_signal(s, loop)))

    try:
        # The logic below of handling the writer worker task separately is to
        # ensure we can terminate properly when all the other tasks have
        # finished as in the case of using file input instead of SSH
        svc_tasks = [svc.run() for svc in svcs]
        tasks = [nodes[node].run() for node in nodes]
        tasks += [run_output_worker(queue, outputs, logger)]
        tasks += svc_tasks

        if not userargs.no_coalescer:
            tasks += [start_and_monitor_coalescer(userargs.config, cfg,
                                                  logger)]
        while tasks:
            try:
                done, pending = await asyncio.wait(
                    tasks, return_when=asyncio.FIRST_COMPLETED)
                tasks = list(pending)
                if tasks and any(i._coro in svc_tasks for i in tasks):
                    continue
                else:
                    break
            except asyncio.CancelledError:
                break
    finally:
        logger.warning("sq-poller: Received terminate signal. Terminating")
        loop.stop()
        return


def poller_main() -> None:

    supported_outputs = ["parquet"]

    parser = argparse.ArgumentParser()
    requiredgrp = parser.add_mutually_exclusive_group(required=True)
    requiredgrp.add_argument(
        "-D",
        "--devices-file",
        type=str,
        help="File with URL of devices to gather data from",
    )
    requiredgrp.add_argument(
        "-a",
        "--ansible-file",
        type=str,
        help="Ansible inventory file of devices to gather data from",
    )
    requiredgrp.add_argument(
        "-i",
        "--input-dir",
        type=str,
        help="Directory where run-once=gather data is"
    )

    parser.add_argument(
        "-n",
        "--namespace",
        type=str, required='--ansible-file' in sys.argv or "-a" in sys.argv,
        help="Namespace to associate for the gathered data"
    )
    parser.add_argument(
        "-o",
        "--outputs",
        nargs="+",
        default=["parquet"],
        choices=supported_outputs,
        type=str,
        help="Output formats to write to: parquet. Use "
        "this option multiple times for more than one output",
    )
    parser.add_argument(
        "-s",
        "--service-only",
        type=str,
        help="Only run this space separated list of services",
    )

    parser.add_argument(
        "-x",
        "--exclude-services",
        type=str,
        help="Exclude running this space separated list of services",
    )

    parser.add_argument(
        "-c",
        "--config",
        type=str, help="alternate config file"
    )

    parser.add_argument(
        "--run-once",
        type=str,
        choices=["gather", "process"],
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--output-dir",
        type=str,
        default=f'{os.path.abspath(os.curdir)}/sqpoller-output',
        help=argparse.SUPPRESS,
    )

    parser.add_argument(
        "--ask-pass",
        default=False,
        action='store_true',
        help="prompt to enter password for login to devices",
    )
    parser.add_argument(
        "--passphrase",
        default=False,
        action='store_true',
        help="prompt to enter private key passphrase",
    )

    parser.add_argument(
        "--envpass",
        default="",
        type=str,
        help="Use named environment variable to retrieve password",
    )

    parser.add_argument(
        "-j",
        "--jump-host",
        default="",
        type=str,
        help="Jump Host via which to access the devices, IP addr or DNS hostname"
    )

    parser.add_argument(
        "-K",
        "--jump-host-key-file",
        default="",
        type=str,
        help="Key file to be used for jump host"
    )

    parser.add_argument(
        "-k",
        "--ignore-known-hosts",
        default=False,
        action='store_true',
        help="Ignore Known Hosts File",
    )

    parser.add_argument(
        "--ssh-config-file",
        type=str,
        default=None,
        help="Path to ssh config file to use. If not set, config file is not used"
    )

    parser.add_argument(
        "--no-coalescer",
        default=False,
        action='store_true',
        help=argparse.SUPPRESS,
    )

    userargs = parser.parse_args()

    if userargs.passphrase:
        userargs.passphrase = getpass.getpass(
            'Passphrase to decode private key file: ')
    else:
        userargs.passphrase = None

    if userargs.ask_pass:
        userargs.ask_pass = getpass.getpass(
            'Password to login to device: ')
    else:
        userargs.ask_pass = None

    if userargs.envpass:
        passwd = os.getenv(userargs.envpass, '')
        if not passwd:
            print(
                f'ERROR: No password in environment variable {userargs.envpass}')
            sys.exit(1)
        userargs.ask_pass = passwd

    uvloop.install()
    cfg = load_sq_config(config_file=userargs.config)
    if not cfg:
        print("Could not load config file, aborting")
        sys.exit(1)

    try:
        asyncio.run(start_poller(userargs, cfg))
    except (KeyboardInterrupt, RuntimeError):
        pass
    except Exception:
        import traceback
        traceback.print_exc()

    sys.exit(0)


if __name__ == '__main__':
    poller_main()



from suzieq.poller.services.service import Service


class TopCpuService(Service):

    def _common_data_cleaner(self, processed_data, raw_data):

        for entry in processed_data:
            for i in ["virtualMem", "residentMem", "cacheMem", "usedMem",
                      "totalMem", "freeMem"]:
                try:
                    entry[i] = int(entry[i])
                except ValueError:
                    if entry[i].endswith("g"):
                        val = float(entry[i].split("g")[0])*1000000
                        entry[i] = int(val)

        return processed_data



import time
from nubia import command, argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.topcpu import TopcpuObj


@command("topcpu", help="Act on topcpu data")
class TopcpuCmd(SqCommand):
    """Information about Top CPU users"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            sqobj=TopcpuObj
        )

    @command("show")
    def show(self):
        """Show topcpu info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self.sqobj.get(
            hostname=self.hostname, columns=self.columns,
            namespace=self.namespace
        )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)



from suzieq.sqobjects.basicobj import SqObject


class SqPollerObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='sqPoller', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'service',
                                'status', 'query_str']
        self._valid_arg_vals = {
            'status': ['all', 'pass', 'fail'],
        }



from .parquetdb import SqParquetDB
import logging


def get_sqdb(cfg: dict, logger: logging.Logger) -> SqParquetDB:
    """Return the parquet db engine for the appropriate table

    :param cfg: dict, Suzieq configuration dictionary
    :param logger: logging.Logger, logger to hook into for logging, can be None
    :returns: a parquet database engine to use for read/wr, other DB ops
    :rtype: SqParquetDB

    """
    return SqParquetDB(cfg, logger)


__all__ = [get_sqdb]



from suzieq.sqobjects.basicobj import SqObject


class TablesObj(SqObject):

    def __init__(self, **kwargs) -> None:
        # We're passing any table name to get init to work
        super().__init__(table='tables', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns',
                                'query_str']



import os
from typing import List
import asyncio
from datetime import datetime, timezone
import time
import copy
import logging
import json
import yaml
from tempfile import mkstemp
from dataclasses import dataclass, field
from http import HTTPStatus
from collections import defaultdict

import pyarrow as pa

from suzieq.poller.services.svcparser import cons_recs_from_json_template
from suzieq.utils import known_devtypes
from suzieq.version import SUZIEQ_VERSION

HOLD_TIME_IN_MSECS = 60000  # How long b4 declaring node dead


@dataclass
class RsltToken:
    start_time: int    # When this cmd was first posted to node
    nodename: str      # Name of node, used to get poller cb again
    bootupTimestamp: int
    nodeQsize: int        # Size of the node q at queuing time
    service: str          # Name of this service, if node needs it
    timeout: int          # timeout value for cmd to complete


@dataclass
class ServiceStats:
    total_time: List[float] = field(default_factory=list)
    gather_time: List[float] = field(default_factory=list)
    svcQsize: List[float] = field(default_factory=list)
    nodeQsize: List[float] = field(default_factory=list)
    wrQsize: List[float] = field(default_factory=list)
    rxBytes: List[float] = field(default_factory=list)
    empty_count: int = 0
    time_excd_count: int = 0    # Number of times total_time > poll period
    next_update_time: int = 0   # When results will be logged


class Service(object):

    def get_poller_schema(self):
        return self._poller_schema

    def set_poller_schema(self, value: dict):
        self._poller_schema = value

    def __init__(self, name, defn, period, stype, keys, ignore_fields, schema,
                 queue, run_once="forever"):
        self.name = name
        self.defn = defn
        self.ignore_fields = ignore_fields or []
        self.writer_queue = queue
        self.keys = keys
        self.schema = schema.get_arrow_schema()
        self.period = period
        self.stype = stype
        self.logger = logging.getLogger(__name__)
        self.run_once = run_once
        self.post_timeout = 5
        self.sigend = False
        self.version = schema.version

        self.update_nodes = False  # we have a new node list
        self.rebuild_nodelist = False  # used only when a node gets init
        self.node_postcall_list = {}
        self.new_node_postcall_list = {}
        self.previous_results = {}
        self._poller_schema = {}
        self.node_boot_times = defaultdict(int)
        self._failed_node_set = set()

        self.poller_schema = property(
            self.get_poller_schema, self.set_poller_schema)

        # The queue to which all nodes will post the result of the command
        self.result_queue = asyncio.Queue()

        # Add the hidden fields to ignore_fields
        self.ignore_fields.append("timestamp")

        if "namespace" not in self.keys:
            self.keys.insert(0, "namespace")

        if "hostname" not in self.keys:
            self.keys.insert(1, "hostname")

        self.partition_cols = schema.get_partition_columns()

        # Setup dictionary of NOS specific extracted data cleaners
        self.dev_clean_fn = {}
        common_dev_clean_fn = getattr(self, '_common_data_cleaner', None)
        for x in known_devtypes():
            if x.startswith('junos'):
                dev = 'junos'
            else:
                dev = x
            self.dev_clean_fn[x] = getattr(
                self, f'_clean_{dev}_data', None) or common_dev_clean_fn

    @ staticmethod
    def is_status_ok(status: int) -> bool:
        if status == 0 or status == HTTPStatus.OK:
            return True
        return False

    def get_data(self):
        """provide the data that is interesting for a service

        does not include queue or schema"""

        r = {}
        for field in 'name defn ignore_fields keys period stype partition_cols'.split(' '):
            r[field] = getattr(self, field)
        # textFSM objects can't be jsoned, so changing it
        for device in r['defn']:
            if 'textfsm' in r['defn'][device]:
                r['defn'][device]['textfsm'] = str(
                    r['defn'][device]['textfsm'])
        return r

    def set_nodes(self, node_call_list):
        """New node list for this service"""
        if self.node_postcall_list:
            self.new_node_postcall_list = node_call_list
            self.update_nodes = True
        else:
            self.node_postcall_list = node_call_list

    def _get_default_vals(self) -> dict:
        return({
            pa.string(): "",
            pa.int32(): 0,
            pa.int64(): 0,
            pa.float32(): 0.0,
            pa.float64(): 0.0,
            pa.date64(): 0.0,
            pa.bool_(): False,
            pa.list_(pa.string()): [],
            pa.list_(pa.int64()): [],
        })

    def get_empty_record(self):
        map_defaults = self._get_default_vals()

        defaults = [map_defaults[x] for x in self.schema.types]
        rec = dict(zip(self.schema.names, defaults))

        return rec

    def _dump_output(self, outputs) -> str:
        """Dump the output to a YAML file"""
        if outputs:
            yml = {"service": self.name,
                   "type": self.run_once,
                   "output": outputs}
            fd, name = mkstemp(suffix=f"-{self.name}-poller.yml")
            f = os.fdopen(fd, "w")
            f.write(yaml.dump(yml))
            f.close()
            return name
        return ""

    def get_diff(self, old, new):
        """Compare list of dictionaries ignoring certain fields
        Return list of adds and deletes
        """
        adds = []
        dels = []
        koldvals = {}
        knewvals = {}
        koldkeys = []
        knewkeys = []

        # keys that start with _ are transient and must be ignored
        # from comparison
        for i, elem in enumerate(old):
            vals = [v for k, v in elem.items()
                    if k not in self.ignore_fields and not k.startswith('_')]
            kvals = [v for k, v in elem.items() if k in self.keys]
            koldvals.update({tuple(str(vals)): i})
            koldkeys.append(kvals)

        for i, elem in enumerate(new):
            vals = [v for k, v in elem.items()
                    if k not in self.ignore_fields and not k.startswith('_')]
            kvals = [v for k, v in elem.items() if k in self.keys]
            knewvals.update({tuple(str(vals)): i})
            knewkeys.append(kvals)

        addlist = [v for k, v in knewvals.items() if k not in koldvals.keys()]
        dellist = [v for k, v in koldvals.items() if k not in knewvals.keys()]

        adds = [new[v] for v in addlist]
        dels = [old[v] for v in dellist if koldkeys[v] not in knewkeys]

        if adds and self.stype == "counters":
            # If there's a change in any field of the counters, update them all
            # simplifies querying
            adds = new

        return adds, dels

    def textfsm_data(self, raw_input, fsm_template, entry_type, schema, data):
        """Convert unstructured output to structured output"""

        records = []
        fsm_template.Reset()
        res = fsm_template.ParseText(raw_input)

        for entry in res:
            metent = dict(zip(fsm_template.header, entry))
            if entry_type:
                metent['_entryType'] = entry_type
            records.append(metent)

        return records

    async def post_results(self, result, token) -> None:
        """The callback that nodes use to post the results back to the service"""
        if self.result_queue:
            self.result_queue.put_nowait((token, result))
        else:
            self.logger.error(f"No queue for service {self.name}")

    def call_node_postcmd(self, postcall, nodename) -> None:
        """Start data gathering by calling the post command list"""
        if postcall and postcall['postq']:
            token = RsltToken(int(time.time()*1000), nodename, 0, 0, self.name,
                              self.period-5)
            postcall['postq'](self.post_results, self.defn, token)

    def clean_json_input(self, data):
        """Clean the JSON input data that is sometimes messed up
        Each service can implement its own version of the cleanup
        """
        return None

    async def start_data_gather(self) -> None:
        """Start data gathering by calling the post command list
        This is only used to fire up the calls the first time. After this,
        each node gets into a self-adapting curve of calling at least after
        the duration specified for the service period. If nodes are slower,
        they'll be scheduled to run after the specified service period after
        their return. We could use this to track slow nodes.
        """

        for node in self.node_postcall_list:
            try:
                self.call_node_postcmd(self.node_postcall_list[node],
                                       node),
            except TimeoutError:
                self.logger.warning(f"Node post failed for {node}")
                continue

    def _process_each_output(self, elem_num, data):
        """Workhorse processing routine for each element in output"""

        result = []
        if (Service.is_status_ok(data.get('status', -1))):
            if not data["data"]:
                return result

            # Check host-specific normalization if any
            nfn = self.defn.get(data.get("hostname"), None)
            if not nfn:
                nfn = self.defn.get(data.get("devtype"), None)
            if nfn:
                # If we're riding on the coattails of another device
                # get that device's normalization function
                copynfn = nfn.get("copy", None)
                if copynfn:
                    nfn = self.defn.get(copynfn, {})
                norm_str = nfn.get("normalize", None)
                if norm_str is None and isinstance(nfn.get('command', None),
                                                   list):
                    # Can happen because we've a list of commands associated
                    # Pick the normalization function associated with this
                    # output
                    norm_str = nfn.get('command', [])[elem_num].get(
                        'normalize', None)
                if norm_str:
                    if isinstance(data["data"], str):
                        try:
                            in_info = json.loads(data["data"])
                        except json.JSONDecodeError:
                            in_info = self.clean_json_input(data)
                            if not in_info:
                                self.logger.error(
                                    "Received non-JSON output where "
                                    "JSON was expected for {} on "
                                    "node {}, {}".format(
                                        data["cmd"], data["hostname"],
                                        data["data"]
                                    )
                                )
                                return result
                            try:
                                in_info = json.loads(in_info)
                            except json.JSONDecodeError:
                                self.logger.error(
                                    "Received non-JSON output where "
                                    "JSON was expected for {} on "
                                    "node {}, {}".format(
                                        data["cmd"], data["hostname"],
                                        data["data"]
                                    )
                                )
                                return result
                    else:
                        in_info = data["data"]

                    if in_info:
                        # EOS' HTTP returns errors like this.
                        if isinstance(in_info, dict):
                            tmp = in_info.get('data', [])
                            if tmp and tmp[0].get('errors', {}):
                                return []

                        result = cons_recs_from_json_template(
                            norm_str, in_info)

                else:
                    tfsm_template = nfn.get("textfsm", None)
                    entry_type = None
                    if tfsm_template is None and isinstance(
                            nfn.get('command', None), list):
                        # Can happen because we've a list of cmds associated
                        # Pick the normalization function associated with this
                        # output
                        tfsm_template = nfn.get('command', [])[elem_num].get(
                            'textfsm', None)
                        entry_type = nfn.get('command', [])[elem_num].get(
                            '_entryType', None)

                    if not tfsm_template:
                        return result

                    in_info = []
                    if isinstance(data['data'], dict):
                        if "output" in data["data"]:
                            in_info = data["data"]["output"]
                        elif "messages" in data["data"]:
                            # This is Arista's bash output format
                            in_info = data["data"]["messages"][0]
                    else:
                        in_info = data['data']
                    if not in_info:
                        return result  # No data to process
                    # Clean data is invoked inside this due to the way we
                    # munge the data and force the types to adhere to the
                    # specified type
                    result = self.textfsm_data(
                        in_info, tfsm_template, entry_type, self.schema, data
                    )
            else:
                self.logger.error(
                    "{}: No normalization/textfsm function for device {}"
                    .format(self.name, data["hostname"]))

        return result

    def process_data(self, data):
        """Derive the data to be stored from the raw input"""
        result_list = []
        for i, item in enumerate(data):
            tmpres = self._process_each_output(i, item)
            result_list.append(tmpres)

        result = self.merge_results(result_list, data)
        return self.clean_data(result, data)

    def get_key_flds(self):
        """Get the key fields associated with this service. 
        Its a function because we want to override it.
        """
        return list(filter(lambda x: x not in ['namespace', 'hostname'],
                           self.keys))

    def merge_results(self, result_list, data):
        int_res = {}
        keyflds = self.get_key_flds()

        for result in result_list:
            for entry in result:
                keyvals = []
                for kfld in keyflds:
                    if not entry.get(kfld, None):
                        keyvals.append('_default')
                    else:
                        val = entry.get(kfld, '')
                        if not isinstance(val, str):
                            val = str(val)
                        keyvals.append(val)

                if entry.get('_entryType', ''):
                    keyvals.append(entry['_entryType'])

                key = '-'.join(keyvals)

                if key not in int_res:
                    int_res[key] = entry
                else:
                    existing = int_res[key]
                    for fld in entry:
                        if entry[fld] and not existing.get(fld, None):
                            existing[fld] = entry[fld]
                        elif entry[fld]:
                            # cleanup routine specific to service can resolve
                            existing[fld+"-_2nd"] = entry[fld]

        final_res = list(int_res.values())

        return(final_res)

    def _get_devtype_from_input(self, input):
        if isinstance(input, list):
            return input[0].get('devtype', None)
        else:
            return input.get('devtype', None)

    def clean_data(self, processed_data, raw_data):
        """Massage the extracted data to match record specified by schema"""

        dev_clean_fn = self.dev_clean_fn.get(self._get_devtype_from_input(
            raw_data), None)
        if dev_clean_fn:
            processed_data = dev_clean_fn(processed_data, raw_data)

        return self.clean_data_common(processed_data, raw_data)

    def clean_data_common(self, processed_data, raw_data):
        """Fix the type and default value of of each extracted field

        This routine is common to all services. It ensures that all the missing
        fields, as defined by the schema, are added to the records extracted.
        Furthermore, each field is set to the specified type.
        """

        # Build default data structure
        schema_rec = {}
        def_vals = self._get_default_vals()

        ptype_map = {
            pa.string(): str,
            pa.int32(): int,
            pa.int64(): int,
            pa.float32(): float,
            pa.float64(): float,
            pa.date64(): float,
            pa.list_(pa.string()): list,
            pa.list_(pa.int64()): list,
            pa.bool_(): bool,
        }

        for field in self.schema:
            default = def_vals[field.type]
            schema_rec.update({field.name: default})

        if isinstance(raw_data, list):
            read_from = raw_data[0]
        else:
            read_from = raw_data
        for entry in processed_data or []:
            entry.update({"hostname": read_from["hostname"]})
            entry.update({"namespace": read_from["namespace"]})
            entry.update({"timestamp": read_from["timestamp"]})
            entry.update({"sqvers": self.version})
            for fld in schema_rec:
                if fld not in entry:
                    if fld == "active":
                        entry.update({fld: True})
                    else:
                        entry.update({fld: schema_rec[fld]})
                else:
                    fld_type = self.schema.field(fld).type
                    if not isinstance(entry[fld], ptype_map[fld_type]):
                        try:
                            entry[fld] = ptype_map[fld_type](entry[fld])
                        except (ValueError, TypeError):
                            entry[fld] = schema_rec[fld]
                    elif isinstance(entry[fld], list):
                        for i, ele in enumerate(entry[fld]):
                            if not isinstance(ele,
                                              ptype_map[fld_type.value_type]):
                                try:
                                    if ptype_map[fld_type.value_type] == int:
                                        entry[fld][i] = int(entry[fld][i])
                                    elif ptype_map[fld_type.value_type] == str:
                                        entry[fld][i] = str(entry[fld][i])
                                    else:
                                        raise ValueError
                                except (ValueError, TypeError):
                                    entry[fld][i] = schema_rec[fld]
        return processed_data

    async def commit_data(self, result, namespace, hostname):
        """Write the result data out"""
        records = []
        prev_res = self.previous_results.get(hostname, [])

        if result or prev_res:
            adds, dels = self.get_diff(prev_res, result)
            if adds or dels:
                self.previous_results[hostname] = copy.deepcopy(result)
                for entry in adds:
                    records.append(entry)
                for entry in dels:
                    if entry.get("active", True):
                        # If there's already an entry marked as deleted
                        # No point in adding one more
                        entry.update({"active": False})
                        entry.update(
                            {"timestamp":
                             int(datetime.now(tz=timezone.utc).timestamp() * 1000)}
                        )
                        records.append(entry)

                self._post_work_to_writer(records)

    def _post_work_to_writer(self, records: dict):
        """This posts the data to be written to the worker queue"""
        if records:
            self.writer_queue.put_nowait(
                {
                    "records": records,
                    "topic": self.name,
                    "schema": self.schema,
                    "partition_cols": self.partition_cols
                }
            )

    def compute_basic_stats(self, statsList, newval: int) -> None:
        """Compute min/max/avg for given stats with the new value

        :param statsList: list consisting of 3 values in order: min, max, avg
        :type statsList: list_

        :param newval: The newval to update the list with
        :type newval: int
        """
        if not statsList:
            statsList = [newval, newval, newval]
            return statsList

        if newval < statsList[0]:
            statsList[0] = newval
        if newval > statsList[1]:
            statsList[1] = newval
        statsList[2] = (statsList[2]+newval)/2
        if statsList[2] < 0.001:
            statsList[2] = 0

        return statsList

    def update_stats(self, stats: ServiceStats, total_time: int,
                     gather_time: int, qsize: int, wrQsize: int,
                     nodeQsize: int, rxBytes) -> bool:
        """Update per-node stats"""
        write_stat = False
        now = int(time.time()*1000)

        stats.total_time = self.compute_basic_stats(stats.total_time,
                                                    total_time)
        stats.gather_time = self.compute_basic_stats(stats.gather_time,
                                                     gather_time)
        stats.svcQsize = self.compute_basic_stats(stats.svcQsize, qsize)
        stats.wrQsize = self.compute_basic_stats(stats.wrQsize, wrQsize)
        stats.nodeQsize = self.compute_basic_stats(stats.nodeQsize, nodeQsize)
        stats.rxBytes = self.compute_basic_stats(stats.rxBytes, rxBytes)

        if total_time > self.period*1000:
            stats.time_excd_count += 1
            write_stat = True

        if now > stats.next_update_time:
            stats.next_update_time = now + 5*60*1000  # 5 mins
            write_stat = True

        return write_stat

    async def run(self):
        """Start the service"""
        self.logger.info(f"running service {self.name} ")

        if self.run_once:
            total_nodes = len(self.node_postcall_list)
        else:
            total_nodes = 0
        # Fire up the initial posts
        await self.start_data_gather()
        loop = asyncio.get_running_loop()
        pernode_stats = defaultdict(lambda: ServiceStats())

        while True:
            try:
                token, output = await self.result_queue.get()
            except asyncio.CancelledError:
                self.logger.warning(
                    f'Service: {self.name}: Received signal to terminate')
                return
            qsize = self.result_queue.qsize()

            self.logger.debug(f"Extracted response for {self.name} service")
            if isinstance(token, list):
                token = token[0]
            gather_time = int(time.time()*1000) - token.start_time

            status = HTTPStatus.NO_CONTENT        # Empty content
            write_poller_stat = False
            rxBytes = 0

            if output:
                ostatus = [x.get('status', -1) for x in output]

                write_poller_stat = not all([Service.is_status_ok(x)
                                             for x in ostatus])
                status = ostatus[0]
                if (status in [0, 200]):
                    rxBytes = len(output.__str__())

                # We don't expect the output from two different hostnames
                nodename = output[0]["hostname"]
                self.logger.debug(f"Extracted response from {nodename} for "
                                  f"{self.name} service")
                # We can terminate processing if we've no data returned
                # and we're reading inputs from a file
                if nodename == '_filedata' and status == HTTPStatus.NO_CONTENT:
                    return

                # Process data independent of status to return an empty list
                # so that the output can be updated. For example, if a service
                #  is disabled, this can be the condition.
                if self.run_once == "gather":
                    if self.is_status_ok(status):
                        self._post_work_to_writer(json.dumps(output, indent=4))
                    total_nodes -= 1
                    if total_nodes <= 0:
                        self.logger.warning(
                            f'Service: {self.name}: Finished gathering data')
                        return
                    continue

                try:
                    result = self.process_data(output)
                except Exception:
                    result = []
                    status = HTTPStatus.BAD_GATEWAY
                    write_poller_stat = True
                    self.logger.exception(
                        f'Processing data failed for service '
                        f'{self.name} on node {nodename}')

                # Don't write the error every time the failure happens
                if write_poller_stat:
                    if nodename in self._failed_node_set:
                        write_poller_stat = False
                    else:
                        self._failed_node_set.add(nodename)
                elif nodename in self._failed_node_set:
                    # So there was no error in this command that had failed b4
                    self._failed_node_set.remove(nodename)

                # If a node from init state to good state, hostname will change
                # So fix that in the node list
                hostname = output[0]["hostname"]
                if self.run_once == "process":
                    if self.is_status_ok(status):
                        self._post_work_to_writer(json.dumps(result, indent=4))
                    total_nodes -= 1
                    if total_nodes <= 0:
                        return
                    continue

                # Check if a node has rebooted
                if token.bootupTimestamp != self.node_boot_times[hostname]:
                    self.node_boot_times[hostname] = token.bootupTimestamp
                    # Flush the prev results data for this node
                    self.previous_results[hostname] = []
                await self.commit_data(result, output[0]["namespace"],
                                       hostname)
            else:
                if self.run_once == "gather" or self.run_once == "process":
                    total_nodes -= 1
                    if total_nodes <= 0:
                        self.logger.info(
                            f'Service: {self.name}: Finished gathering data')
                        return
                    continue

            total_time = int(time.time()*1000) - token.start_time

            if output:
                statskey = output[0]["namespace"] + '/' + output[0]["hostname"]

                stats = pernode_stats[statskey]
                write_poller_stat = (self.update_stats(
                    stats, total_time, gather_time, qsize,
                    self.writer_queue.qsize(), token.nodeQsize, rxBytes) or
                    write_poller_stat)
                pernode_stats[statskey] = stats
                if write_poller_stat:
                    poller_stat = [
                        {"hostname": output[0]["hostname"] or output[0]['address'],
                         "sqvers": self.poller_schema_version,
                         "namespace": output[0]["namespace"],
                         "active": True,
                         "service": self.name,
                         "status": status,
                         "svcQsize": stats.svcQsize,
                         "wrQsize": stats.wrQsize,
                         "nodeQsize": stats.nodeQsize,
                         "rxBytes": stats.rxBytes,
                         "pollExcdPeriodCount": stats.time_excd_count,
                         "gatherTime": stats.gather_time,
                         "totalTime": stats.total_time,
                         "version": SUZIEQ_VERSION,
                         "nodesPolledCnt": len(self.node_postcall_list),
                         "nodesFailedCnt": len(self._failed_node_set),
                         "timestamp": int(datetime.now(tz=timezone.utc)
                                          .timestamp() * 1000)}]

                    self.writer_queue.put_nowait(
                        {
                            "records": poller_stat,
                            "topic": "sqPoller",
                            "schema": self.poller_schema,
                            "partition_cols": self.partition_cols
                        })

            # Post a cmd to fire up the next poll after the specified period
            self.logger.debug(
                f"Rescheduling service for {self.name} service")
            loop.call_later(self.period, self.call_node_postcmd,
                            self.node_postcall_list.get(token.nodename),
                            token.nodename)



import time

from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.tables import TablesObj


@command("table", help="Information about the various tables")
class TableCmd(SqCommand):
    """Meta information about the various data gathered"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            sqobj=TablesObj,
        )

    @command("show")
    def show(self, **kwargs):
        """Show meta-information about various tables gathered by Suzieq
        """
        now = time.time()
        df = self.sqobj.get(hostname=self.hostname, namespace=self.namespace,
                            columns=self.columns)
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @ command("describe")
    @ argument("table", description="interface name to qualify")
    def describe(self, table: str = "tables", **kwargs):
        """
        Describe fields in table
        """

        if not table:
            df = pd.DataFrame({'error': ['ERROR: Must specify a table']})
            return self._gen_output(df)

        if self.columns != ['default']:
            df = pd.DataFrame(
                {'error': ['ERROR: Cannot specify columns for command']})
            return self._gen_output(df)

        now = time.time()
        df = self.sqobj.describe(table=table)
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._gen_output(df, dont_strip_cols=True)



#!/usr/bin/env python

import sys
import typing
import pyarrow.parquet as pq
import pyarrow as pa
from pathlib import Path, PosixPath
from suzieq.utils import Schema, SchemaForTable
import concurrent.futures
import logging


def convert_file(file: PosixPath, output_dir: str, schema: SchemaForTable,
                 defaults: typing.List,
                 arrow_schema: pa.lib.Schema,
                 new_partition_columns=['sqvers', 'namespace', 'hostname']):
    """Read the dataset from file and convert it. No compaction."""

    logger.info(f'Reading {file}')
    df = pq.ParquetDataset(file, use_legacy_dataset=True,
                           validate_schema=False, partitioning='hive') \
        .read() \
        .to_pandas()

    part_data = [x.split('=') for x in file.parts if '=' in x]
    for elem in part_data:
        df[elem[0]] = elem[1]

    for column in filter(lambda x: x['name'] not in df.columns, schema):
        df[column['name']] = column.get('default', defaults[column['type']])

    # If there's the original ifname saved up, then eliminate this unnecessary
    # field as this model is no longer necessary

    if 'origIfname' in df.columns:
        if 'ifname' in df.columns:
            df = df.drop(columns=['ifname']) \
                   .rename(columns={'origIfname': 'ifname'})
        elif 'oif' in df.columns:
            df = df.drop(columns=['oif']) \
                   .rename(columns={'origIfname': 'oif'})

    # convert all dtypes to whatever is desired
    for column in df.columns:
        df[column] = df[column].astype(arrow_schema.field(column)
                                       .type.to_pandas_dtype())

    table = pa.Table.from_pandas(df, schema=arrow_schema, preserve_index=False)

    pq.write_to_dataset(
        table,
        root_path=output_dir,
        partition_cols=new_partition_columns,
        version="2.0",
        compression='ZSTD',
        row_group_size=100000,
    )
    logger.info(f'Wrote converted {file}')

    return file


def convert_dir(input_dir: str, output_dir: str, sqschema: typing.List,
                arrow_schema: pa.lib.Schema):
    """Spawn off threads to convert directory"""

    files = Path(input_dir).glob('**/*.parquet')

    defaults = {
        pa.string(): "",
        pa.int32(): 0,
        pa.int64(): 0,
        pa.float32(): 0.0,
        pa.float64(): 0.0,
        pa.date64(): 0.0,
        pa.bool_(): False,
        pa.list_(pa.string()): ['-'],
        pa.list_(pa.int64()): [],
    }

    with concurrent.futures.ProcessPoolExecutor(max_workers=None) as thread:
        threads = {thread.submit(convert_file, item,
                                 output_dir, sqschema, defaults, arrow_schema)
                   for item in files}
        for future in concurrent.futures.as_completed(threads):
            try:
                _ = future.result()
            except Exception:
                logger.exception(f'Exception occcurred with {future}')


if __name__ == "__main__":
    if len(sys.argv) < 4:
        print('Usage: convert_parquet <input dir> <output_dir> <schema_dir>')
        sys.exit(1)

    input_dir = Path(sys.argv[1])
    output_dir = sys.argv[2]
    schemas = Schema(sys.argv[3])
    service = input_dir.parts[-1]
    svc_schema = SchemaForTable(service, schema=schemas)
    arrow_schema = svc_schema.get_arrow_schema()
    sqschema = svc_schema.get_raw_schema()

    logging.basicConfig(stream=sys.stdout, level=logging.WARNING)
    logger = logging.getLogger('sq-converter')
    convert_dir(input_dir, output_dir, sqschema, arrow_schema)



import logging
from typing import List
from importlib import import_module
import inspect

from suzieq.exceptions import DBNotFoundError


def get_sqdb_engine(cfg: dict, table_name: str, dbname: str,
                    logger: logging.Logger):
    """Return the appropriate DB reader for the given table

    This function is dumb right now. Will get smarter as we add support for
    more database readers.

    :param cfg: dict, Suzieq loaded config
    :param table_name: str, Name of the table you want to read
    :param dbname: str, specify the DB engine you want to get handle for
                   can be empty
    :param logger: logging.logger, the logger to use for logging
    :returns: the class that is the reader for this object
    :rtype:
    """
    # Load the available engines
    avail_db = ['parquet']

    use_db = None
    if table_name and table_name in cfg.get('db', {}):
        use_db = cfg['db'][table_name]
    elif dbname and dbname in avail_db:
        use_db = dbname
    else:
        use_db = 'parquet'      # The default

    if not use_db:
        logger.error(f'Unable to find a DB for {table_name}/{dbname}')
        return None

    try:
        eng_mod = import_module(f'suzieq.db.{use_db}')
    except ModuleNotFoundError:
        return None

    for mbr in inspect.getmembers(eng_mod):
        if mbr[0] == 'get_sqdb' and inspect.isfunction(mbr[1]):
            return mbr[1](cfg, logger)

    return None


def do_coalesce(cfg: dict, tables: List[str], period: str = '1h',
                logger: logging.Logger = None,
                no_sqpoller: bool = False) -> None:
    """The main coalescer routine, can run once or periodically.

    It calls the DB-specific coalescer. The period is necessary to pass to the
    DB coalescer even in case of run_once. By default the period is '1h'.

    :param cfg: dict, the SUzieq config dictionary
    :param tables: List[str], the list of tables to coalesce
    :param period: str, the string of how periodically the poller runs,
                   Examples are '1h', '1d' etc.
    :param run_once: bool, If true, run once and exit
    :param logger, logging.Logger, logger to use
    :param no_sqpoller: bool, ignore sqpoller
    :returns: dictionary of stats about coalescing
    :rtype: None

    """

    if not logger:
        logger = logging.getLogger()

    dbeng = get_sqdb_engine(cfg, None, 'parquet', logger)
    if not dbeng:
        logger.error('Unable to get DB object for DB parquet')
        raise DBNotFoundError

    return dbeng.coalesce(tables, period, no_sqpoller)



from abc import ABC, abstractmethod


class SqEngineObj(ABC):
    '''Interface class for adding analyzer engine to Suzieq'''

    def __init__(self):
        pass

    @abstractmethod
    def get(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def summarize(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def unique(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def aver(self, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def top(self, **kwargs):
        raise NotImplementedError



import re
import numpy as np
from suzieq.poller.services.service import Service
from suzieq.utils import expand_nxos_ifname


class InventoryService(Service):
    """Inventory service"""

    def _clean_data_common(self, processed_data, raw_data):
        return processed_data

    def _clean_eos_data(self, processed_data, raw_data):
        new_data = []
        for entry in processed_data:
            # Because of the JSON structure of the data, we could not
            # easily extract the data.
            for portnum in entry.get('_xcvrSlots', []):
                model = entry['_xcvrSlots'][portnum].get('modelName', '')
                if model:
                    status = 'present'
                else:
                    status = 'absent'
                newnentry = {
                    "name": f"port-{portnum}",  # Arista adds a /1 to ifname
                    "model": model,
                    "type": "xcvr",
                    "status": status,
                    "partNum": "",
                    "version": entry['_xcvrSlots'][portnum].get('hardwareRev',
                                                                ''),
                    "vendor": entry['_xcvrSlots'][portnum].get('mfgName',
                                                               '').lower(),
                    "serial": entry['_xcvrSlots'][portnum].get('serialNum',
                                                               ''),
                }
                new_data.append(newnentry)

            for fanNum in entry.get('_fanTraySlots', []):
                serial = entry['_fanTraySlots'][fanNum].get('serialNum', '')
                if serial == 'N/A':
                    serial = ''
                newentry = {
                    "name": f"fan-{fanNum}",
                    "type": "fan",
                    "serial": serial,
                    "model": entry['_fanTraySlots'][fanNum]
                    .get('name', ''),
                    "status": "present",
                    "vendor": entry['_fanTraySlots'][fanNum].get('vendor', 'Arista'),
                    "version": ''
                }
                new_data.append(newentry)

            for powerNum in entry.get('_powerSupplySlots', []):
                newentry = {
                    "name": f"power-{powerNum}",
                    "type": "power",
                    "serial": entry['_powerSupplySlots'][powerNum]
                    .get('serialNum', ''),
                    "model": entry['_powerSupplySlots'][powerNum]
                    .get('name', ''),
                    "status": "present",
                    "vendor": entry['_powerSupplySlots'][powerNum].get('vendor', 'Arista'),
                    "version": ''  # Arista does not provide this
                }
                new_data.append(newentry)

            for fabricNum in entry.get('_cardSlots', []):
                model = entry['_cardSlots'][fabricNum].get('modelName', '')
                if model == "Not Inserted":
                    model = ''
                    status = 'absent'
                else:
                    status = 'present'
                ftype, num, _ = re.split(r'(\d)$', fabricNum)
                ftype = ftype.lower()
                newentry = {
                    "name": f'{ftype}-{num}',
                    "type": ftype,
                    "serial": entry['_cardSlots'][fabricNum]
                    .get('serialNum', ''),
                    "model": model.lower(),
                    "status": status,
                    "vendor": entry['_cardSlots'][fabricNum].get('vendor', 'Arista'),
                    "version": entry['_cardSlots'][fabricNum].get('hardwareRev', ''),
                }
                new_data.append(newentry)

        return new_data

    def _clean_iosxr_data(self, processed_data, raw_data):
        pass

    def _clean_nxos_data(self, processed_data, raw_data):

        for entry in processed_data:
            name = entry.get('name', '')
            entry['status'] = 'present'
            if name.find('Ethernet') != -1:
                entry['type'] = 'xcvr'
                if entry['_sfp'].startswith("not "):
                    entry['status'] = 'absent'
            elif name.find('Chassis') != -1:
                entry['type'] = 'chassis'
                entry['name'] = name.lower()
                entry['model'] = ' '.join(entry['model'].split()[0:2])
            elif name.find('Slot') != -1:
                entry['type'] = 'linecard'
                entry['name'] = f'linecard-{name.split()[-1]}'
            elif name.find('Power Supply') != -1:
                entry['type'] = 'power'
                entry['name'] = f'power-{name.split()[-1]}'
                entry['model'] = ' '.join(entry['model'].split()[0:2])
            elif name.find('Fan') != -1:
                entry['type'] = 'fan'
                entry['name'] = f'fan-{name.split()[-1]}'
                entry['model'] = ' '.join(entry['model'].split()[0:2])

        return processed_data

    def _clean_junos_data(self, processed_data, raw_data):
        new_data = []
        drop_indices = []

        for i, entry in enumerate(processed_data):
            if not entry['name']:
                drop_indices.append(i)
                continue

            name = entry.get('name', '').lower()
            if name == "midplane":
                entry['type'] = "midplane"
            elif name.startswith("fan"):
                entry['type'] = 'fan'
                if 'tray' in name:
                    entry['name'] = f'fan-{name.split()[0]}'
            elif name.startswith(('pdm', 'pem')):
                entry['type'] = 'power'
                name = name.replace(' ', '-')
            elif name.startswith('cb'):
                entry['type'] = 'mx-scb'
                name = name.replace(' ', '-')
            elif name.startswith('fpc'):
                fabid = name.split()[1]
                entry['type'] = 'linecard'
                name = f'linecard-{fabid}'
            elif name.startswith('routing'):
                entry['type'] = 'supervisor'
                name = name.replace(' ', '-')

            entry['name'] = name
            entry['status'] = 'present'
            entry['vendor'] = 'Juniper'
            submod = entry.get('_chassisSubMod', []) or []
            for ele in submod:
                subname = ele.get('name', [{}])[0].get('data', '')
                if not subname:
                    entry.pop('_chassisSubMod', None)
                    continue
                if not subname.startswith(('PIC', 'MIC')):
                    # Its another entry  away in this list
                    continue
                subid = subname.split()[1]
                port_substr = f'{fabid}/{subid}'
                if subname.startswith('MIC'):
                    nentry = {
                        'name': f'port-adapter-{fabid}/{subid}',
                        'vendor': 'Juniper',
                        'status': 'present',
                        'type': 'port-adapter',
                        'version': ele.get('version', [{}])[0].get('data', ''),
                        'serial': ele.get('serial-number', [{}])[0].get('data', ''),
                        'model': ele.get('model-number', [{}])[0].get('data', ''),
                        'partNum': ele.get('part-number', [{}])[0].get('data', '')
                    }
                    new_data.append(nentry)
                    picent = ele.get('chassis-sub-sub-module', [])
                    for pic in picent:
                        xcvr_ent = pic.get('chassis-sub-sub-sub-module', [])
                        for xent in xcvr_ent:
                            nentry = self._junos_create_xcvr_entry(
                                xent, port_substr)
                            if nentry:
                                new_data.append(nentry)
                else:
                    xcvr_ent = ele.get('chassis-sub-sub-module', [])
                    for xent in xcvr_ent:
                        nentry = self._junos_create_xcvr_entry(
                            xent, port_substr)
                        if nentry:
                            new_data.append(nentry)

            entry.pop('_chassisSubMod', None)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        processed_data.extend(new_data)

        return processed_data

    def _junos_create_xcvr_entry(self, picent: dict, port_substr: str) -> dict:
        xname = picent.get('name', [{}])[0].get('data', '')
        pid = xname.split()[1]
        partnum = picent.get('part-number', [{}])[0].get('data', '')
        if partnum == "NON-JNPR":
            vendor = 'unavailable'
        else:
            vendor = 'Juniper'
        nentry = {
            "name": f'port-{port_substr}/{pid}',
            "type": "xcvr",
            "status": "present",
            'vendor': vendor,
            "version": picent.get('version', [{}])[0].get('data', ''),
            "model": picent.get('description', [{}])[0].get('data', ''),
            "partNum": partnum,
            "serial": picent.get('serial-number', [{}])[0].get(
                'data', ''),
        }

        return nentry

    def _clean_ios_data(self, processed_data, raw_data):
        pass



""" searches the provided parquet-out directory to find any invalid files 
    it just finds all the parquet files and tries to load them as valid files
    if there is an exception, then we know it's a bad file
"""

# TODO
#  figure out how to break the files into multiple processes, otherwise
#  it gets stuck behind a single core

from suzieq.utils import get_latest_files
import argparse
import os
import pyarrow.parquet as pa

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--parquet_dir', type=str, required=True,
                        help="parquet directory")
    userargs = parser.parse_args()

    # this doesn't work, it's broken and doesn't return all the files
    #files = get_latest_files(f"{userargs.parquet_dir}/{table}", view='all')

    all_files = []
    broken_files = []
    for root, dirs, files in os.walk(f"{userargs.parquet_dir}"):
        if not '_archived' in root and not '_broken' in root and not '.sq-coalescer.pid' in files:
            all_files.extend(list(map(lambda x: f"{root}/{x}", files)))
    print(f"{len(all_files)} files")

    
    for file in all_files:
        try:
            parquet_file = pa.ParquetFile(file)
        except pa.lib.ArrowInvalid as e:
            broken_files.append(file)

    print(f"Broken files: {broken_files}")



#!/usr/bin/env python

from typing import List
import sys
import os
import re
import errno
import argparse
import fcntl
from logging import Logger
from time import sleep
from typing import Tuple
from dataclasses import asdict

import pandas as pd

from suzieq.utils import (load_sq_config, Schema, init_logger,
                          SchemaForTable, ensure_single_instance,
                          get_log_params, get_sleep_time)
from suzieq.db import do_coalesce, get_sqdb_engine
from suzieq.version import SUZIEQ_VERSION


def validate_periodstr(periodstr: str) -> Tuple[bool, str]:
    '''Validate the period string specified by user'''

    status = True
    errmsg = ''
    try:
        tm, unit, _ = re.split(r'(\D)', periodstr)

        if unit not in ['m', 'h', 'd', 'w']:
            errmsg = ('Aborting. Invalid unit specified in coalescing period '
                      f'{periodstr}. Allowed values are "m", "h", "d", "w"')
            status = False

        if int(tm) <= 0:
            errmsg = (f'Aborting. Invalid value {tm} specified in '
                      'coalescing period. Value must be > 0')
            status = False
    except ValueError:
        errmsg = ('Aborting. Invalid value specified in coalescing '
                  f'period {periodstr}. Format is <number><m|h|d|w>')
        status = False

    return status, errmsg


def run_coalescer(cfg: dict, tables: List[str], periodstr: str, run_once: bool,
                  logger: Logger, no_sqpoller: bool = False) -> None:
    """Run the coalescer.

    Runs it once and returns or periodically depending on the
    value of run_once. It also writes out the coalescer records
    as a parquet file.

    :param cfg: dict, the Suzieq config file read in
    :param tables: List[str], list of table names to coalesce
    :param periodstr: str, the string of how periodically the poller runs,
                      Examples are '1h', '1d' etc.
    :param run_once: bool, True if you want the poller to run just once
    :param logger: logging.Logger, the logger to write logs to
    :param no_sqpoller: bool, write records even when there's no sqpoller rec
    :returns: Nothing
    :rtype: none

    """

    try:
        schemas = Schema(cfg['schema-directory'])
    except Exception as ex:
        logger.error(f'Aborting. Unable to load schema: {str(ex)}')
        print(f'ERROR: Aborting. Unable to load schema: {str(ex)}')
        sys.exit(1)

    coalescer_schema = SchemaForTable('sqCoalescer', schemas)
    pqdb = get_sqdb_engine(cfg, 'sqCoalescer', None, logger)

    status, errmsg = validate_periodstr(periodstr)
    if not status:
        logger.error(errmsg)
        print(f'ERROR: {errmsg}')
        sys.exit(1)

    while True:
        try:
            stats = do_coalesce(cfg, tables, periodstr, logger, no_sqpoller)
        except Exception:
            logger.exception('Coalescer aborted. Continuing')
        # Write the selftats
        if stats:
            df = pd.DataFrame([asdict(x) for x in stats])
            if not df.empty:
                df['sqvers'] = coalescer_schema.version
                df['version'] = SUZIEQ_VERSION
                df['active'] = True
                df['namespace'] = ''
                pqdb.write('sqCoalescer', 'pandas', df, True,
                           coalescer_schema.get_arrow_schema(), None)

        if run_once:
            break
        sleep_time = get_sleep_time(periodstr)
        sleep(sleep_time)


def coalescer_main():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-s",
        "--service-only",
        type=str,
        help="Only run this space separated list of services",
    )
    parser.add_argument(
        "-x",
        "--exclude-services",
        type=str,
        help="Exclude running this space separated list of services",
    )

    parser.add_argument(
        "-c",
        "--config",
        default=f'{os.getenv("HOME")}/.suzieq/suzieq-cfg.yml',
        type=str, help="alternate config file"
    )
    parser.add_argument(
        "--run-once",
        default=False,
        help='Run the coalescer once and exit',
        action='store_true',
    )
    parser.add_argument(
        "-p",
        "--period",
        type=str,
        help=('Override the period specified in config file with this. '
              'Format is <period><m|h|d|w>. 1h is 1 hour, 2w is 2 weeks etc.')
    )
    parser.add_argument(
        "--no-sqpoller",
        action='store_true',
        help=argparse.SUPPRESS
    )

    userargs = parser.parse_args()

    cfg = load_sq_config(config_file=userargs.config)
    if not cfg:
        print(f'Invalid Suzieq config file {userargs.config}')
        sys.exit(1)

    logfile, loglevel, logsize, log_stdout = get_log_params('coalescer', cfg,
                                                            '/tmp/sq-coalescer.log')
    logger = init_logger('suzieq.coalescer', logfile,
                         loglevel, logsize, log_stdout)

    # Ensure we're the only compacter
    coalesce_dir = cfg.get('coalescer', {})\
        .get('coalesce-directory',
             f'{cfg.get("data-directory")}/coalesced')

    fd = ensure_single_instance(f'{coalesce_dir}/.sq-coalescer.pid',
                                False)
    if not fd:
        print('ERROR: Another coalescer process present')
        logger.error('Another coalescer process present')
        sys.exit(errno.EBUSY)

    timestr = userargs.period or (
        cfg.get('coalescer', {'period': '1h'}).get('period', '1h'))

    schemas = Schema(cfg.get('schema-directory'))
    if userargs.service_only or userargs.exclude_services:
        tables = [x for x in schemas.tables()
                  if (schemas.type_for_table(x) != "derivedRecord")]
        if userargs.service_only:
            tables = [x for x in tables if x in userargs.service_only.split()]
        if userargs.exclude_services:
            tables = [x for x in tables
                      if x not in userargs.exclude_services.split()]
    else:
        tables = []

    run_coalescer(cfg, tables, timestr, userargs.run_once,
                  logger, userargs.no_sqpoller or False)
    os.truncate(fd, 0)
    try:
        fcntl.flock(fd, fcntl.LOCK_UN)
        os.close(fd)
    except OSError:
        pass

    sys.exit(0)


if __name__ == '__main__':
    coalescer_main()



from dataclasses import dataclass
import streamlit as st

import altair as alt
from suzieq.gui.guiutils import gui_get_df
import numpy as np


def get_title():
    return 'Status'


@dataclass
class StatusSessionState:
    namespace: str = ''


def draw_sidebar_status(state, sqobjs):
    '''Draw appropriate sidebar for the page'''

    do_refresh = st.sidebar.button('Refresh')

    devdf = gui_get_df(sqobjs['device'], columns=['namespace', 'hostname'])
    if devdf.empty:
        st.error('Unable to retrieve any namespace info')
        st.stop()

    namespaces = [''] + sorted(devdf.namespace.unique().tolist())
    if state.namespace:
        nsidx = namespaces.index(state.namespace)
    else:
        nsidx = 0
    namespace = st.sidebar.selectbox('Namespace',
                                     namespaces, index=nsidx)

    if do_refresh:
        st.caching.clear_cache()

    st.sidebar.markdown(
        '''This page provides an overview of the overall network status

Select one of the following pages from the Page menu to investigate further.
* __Xplore__: Look at all the data, look at summaries, run asserts, queries and more
* __Path__: Trace the paths between destinations in a given namespace
* __Search__: Search for addresses in various tables. Just type in any address you want to search. You can specify multiple addresses, space separated. See the search page for more help.

__Caching is enabled by default for 90 secs on all pages__. You can clear the cache by hitting the refresh button on this page or selecting "Clear Cache" option from the drop down menu on the top right hand corner
''')

    if namespace != state.namespace:
        state.namespace = namespace


def page_work(state_container):
    '''The main workhorse routine for the XNA page'''

    if 'statusSessionState' not in state_container:
        state_container['statusSessionState'] = StatusSessionState()

    state = state_container.statusSessionState
    draw_sidebar_status(state, state_container.sqobjs)

    col1, mid, col2 = st.columns([2, 1, 2])
    with col1:
        dev_gr = st.empty()
    with col2:
        if_gr = st.empty()
    col3, mid, col4 = st.columns([2, 1, 2])
    with col3:
        bgp_gr = st.empty()
    with col4:
        ospf_gr = st.empty()

    # Get each of the summarize info
    if state.namespace:
        ns = [state.namespace]
    else:
        ns = []
    dev_df = gui_get_df(state_container.sqobjs['device'], namespace=ns,
                        columns=['*'])
    if not dev_df.empty:
        dev_status = dev_df.groupby(by=['namespace', 'status'])['hostname'] \
                           .count() \
                           .reset_index() \
                           .rename({'hostname': 'count'}, axis=1)

        dev_chart = alt.Chart(dev_status, title='Devices') \
                       .mark_bar(tooltip=True) \
                       .encode(y='status', x='count:Q', row='namespace',
                               color=alt.Color(
                                   'status',
                                   scale=alt.Scale(domain=['alive', 'dead', 'neverpoll'],
                                                   range=['green', 'red', 'darkred']))
                               )
        dev_gr.altair_chart(dev_chart)
    else:
        dev_gr.info('No device info found')

    if_df = gui_get_df(state_container.sqobjs['interfaces'],
                       namespace=ns, columns=['*'])
    if not if_df.empty:
        if_df['state'] = np.where((if_df.state == "down") & (
            if_df.adminState == "down"), "adminDown", if_df.state)

        if_status = if_df.groupby(by=['namespace', 'state'])['hostname'] \
                         .count() \
                         .reset_index() \
                         .rename({'hostname': 'count'}, axis=1)

        if_chart = alt.Chart(if_status, title='Interfaces') \
                      .mark_bar(tooltip=True) \
                      .encode(y='state', x='count:Q', row='namespace',
                              color=alt.Color(
                                  'state',
                                  scale=alt.Scale(domain=['up', 'adminDown',
                                                          'down'],
                                                  range=['green', 'orange',
                                                         'red']))
                              )
        if_gr.altair_chart(if_chart)
    else:
        if_gr.info('No Interface info found')

    bgp_df = gui_get_df(state_container.sqobjs['bgp'], namespace=ns,
                        columns=['*'])

    if not bgp_df.empty and ('error' not in bgp_df):
        bgp_status = bgp_df.groupby(by=['namespace', 'state'])['hostname'] \
                           .count() \
                           .reset_index() \
                           .rename({'hostname': 'count'}, axis=1)

        bgp_chart = alt.Chart(bgp_status, title='BGP') \
                       .mark_bar(tooltip=True) \
                       .encode(y='state', x='count:Q', row='namespace',
                               color=alt.Color(
                                   'state',
                                   scale=alt.Scale(
                                       domain=['Established',
                                               'NotEstd', 'dynamic'],
                                       range=['green', 'red', 'orange']))
                               )
        bgp_gr.altair_chart(bgp_chart)

    ospf_df = gui_get_df(state_container.sqobjs['ospf'],
                         namespace=ns, columns=['*'])
    if not ospf_df.empty:
        ospf_df['state'] = np.where(ospf_df.ifState == "adminDown",
                                    "adminDown", ospf_df.adjState)

        ospf_status = ospf_df.groupby(by=['namespace', 'state'])['hostname'] \
                             .count() \
                             .reset_index() \
                             .rename({'hostname': 'count'}, axis=1)

        ospf_chart = alt.Chart(ospf_status, title='OSPF') \
                        .mark_bar(tooltip=True) \
                        .encode(y='state', x='count:Q', row='namespace',
                                color=alt.Color(
                                    'state',
                                    scale=alt.Scale(
                                        domain=['full', 'fail',
                                                'adminDown', 'passive'],
                                        range=['green', 'red', 'orange', 'peach']))
                                )
        ospf_gr.altair_chart(ospf_chart)

    sqdf = gui_get_df(state_container.sqobjs['sqPoller'],
                      columns=['namespace', 'hostname', 'timestamp'],
                      service='device', namespace=ns)
    if not sqdf.empty:
        hosts = sqdf.groupby(by=['namespace'])['hostname'] \
                    .nunique() \
                    .reset_index() \
                    .rename({'hostname': '#devices'}, axis=1)
        times = sqdf.groupby(by=['namespace'])['timestamp'] \
                    .max().reset_index() \
                          .rename({'timestamp': 'lastPolledTime'}, axis=1)
        pstats = times.merge(hosts, on=['namespace'])

        st.subheader('Poller Status')
        st.dataframe(pstats)

    st.experimental_set_query_params(**{})



import time
from nubia import command, argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.macs import MacsObj


@command("mac", help="Act on MAC Table data")
class MacCmd(SqCommand):
    """MAC address table information"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        query_str: str = ' ',
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=MacsObj,
        )

    @command("show")
    @argument("vlan", description="VLAN(s) to qualify output")
    @argument("macaddr",
              description="MAC address(es), in quotes, to qualify output")
    @argument("remoteVtepIp", description="only with this remoteVtepIp; use any for all")
    @argument("bd", description="filter entries with this bridging domain")
    @argument("local", description="filter entries with no remoteVtep")
    @argument("moveCount", description="num of times this MAC has moved")
    def show(self, vlan: str = '', macaddr: str = '', remoteVtepIp: str = '',
             bd: str = '', local: bool = False, moveCount: str = ''):
        """Show MAC table info

        The remoteVtepInfo is set to "-" to allow to fetch local entries only
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        drop_cols = []

        if vlan and (self.columns != ['default'] and self.columns != ['*'] and
                     'vlan' not in self.columns):
            addnl_fields = ['vlan']
            drop_cols.append('vlan')
        else:
            addnl_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                macaddr=macaddr.split(),
                                addnl_fields=addnl_fields,
                                remoteVtepIp=remoteVtepIp.split(),
                                vlan=vlan.split(),
                                localOnly=local,
                                bd=bd,
                                columns=self.columns,
                                namespace=self.namespace,
                                moveCount=moveCount,
                                query_str=self.query_str,
                                )
        if not df.empty and "mackey" in df.columns:
            drop_cols.append('mackey')

        df.drop(columns=drop_cols, inplace=True)

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)



from collections import deque
from dataclasses import dataclass, asdict, field
from typing import List

import pandas as pd
from ipaddress import ip_address
import streamlit as st
from suzieq.utils import convert_macaddr_format_to_colon
from suzieq.gui.guiutils import gui_get_df, SuzieqMainPages


def get_title():
    # suzieq_gui.py has hardcoded this name.
    return 'Search'


@dataclass
class SearchSessionState:
    page: str = SuzieqMainPages.SEARCH
    search_text: str = ''
    past_df = None
    table: str = ''
    namespace: str = ''
    query_str: str = ''
    unique_query: dict = field(default_factory=dict)
    prev_results = deque(maxlen=5)


def build_query(state, search_text: str) -> str:
    '''Build the appropriate query for the search'''

    if not search_text:
        return '', '', []

    unique_query = {}

    addrs = search_text.split()
    if not addrs:
        return '', '', []

    query_str = disjunction = ''
    columns = ['default']

    if addrs[0].startswith('mac'):
        state.table = 'macs'
        addrs = addrs[1:]
    elif addrs[0].startswith('route'):
        state.table = 'routes'
        addrs = addrs[1:]
    elif addrs[0].startswith('arp'):
        state.table = 'arpnd'
        addrs = addrs[1:]
    elif addrs[0].startswith('address'):
        state.table = 'network'
        search_text = ' '.join(addrs[1:])
    elif addrs[0].startswith('vtep'):
        state.table = 'evpnVni'
        if addrs[0] != 'vteps':
            query_str = (f'priVtepIp.isin({addrs[1:]}) or '
                         f'secVtepIp.isin({addrs[1:]})')
            columns = ['namespace', 'hostname', 'priVtepIp',
                       'secVtepIp']
    elif addrs[0].startswith('vni'):
        state.table = 'evpnVni'
        if addrs[0] == 'vnis':
            try:
                vnis = [int(x) for x in addrs[1:]]
            except ValueError:
                vnis = []
            query_str = f'vni.isin({vnis})'
            columns = ['namespace', 'hostname', 'vni']
    elif addrs[0].startswith('asn'):
        state.table = 'bgp'
        if addrs[0] != "asns":
            try:
                asns = [int(x) for x in addrs[1:]]
            except ValueError:
                asns = []
            query_str = f'asn.isin({asns})'
            columns = ['namespace', 'hostname', 'asn']
    elif addrs[0].startswith('vlan'):
        state.table = 'vlan'
        if addrs[0] != "vlans":
            try:
                vlans = [int(x) for x in addrs[1:]]
            except ValueError:
                vlans = []
            query_str = f'vlan.isin({vlans})'
            columns = ['namespace', 'hostname', 'vlan']
    elif addrs[0].startswith('mtus'):
        state.table = 'interface'
        if addrs[0] != "mtus":
            try:
                mtus = [int(x) for x in addrs[1:]]
            except ValueError:
                mtus = []
            query_str = f'mtu.isin({mtus})'
            columns = ['namespace', 'hostname', 'mtu']
    else:
        state.table = 'network'

    if state.table == 'network':
        return search_text, unique_query, columns

    for addr in addrs:
        if addr.lower() == 'vteps':
            unique_query = {'table': 'evpnVni',
                            'column': ['priVtepIp', 'secVtepIp'],
                            'colname': 'vteps'}
        elif addr.lower() == 'vnis':
            unique_query = {'table': 'evpnVni',
                            'column': ['vni'], 'colname': 'vnis'}
        elif addr.lower() == 'asns':
            unique_query = {'table': 'bgp', 'column': ['asn', 'peerAsn'],
                            'colname': 'asns'}
        elif addr.lower() == 'vlans':
            unique_query = {'table': 'vlan', 'column': ['vlan'],
                            'colname': 'vlans'}
        elif addr.lower() == 'mtus':
            unique_query = {'table': 'interfaces', 'column': ['mtu'],
                            'colname': 'mtus'}

        elif '::' in addr:
            if state.table == 'arpnd':
                query_str += f' {disjunction} ipAddress == "{addr}" '
            elif state.table == 'routes':
                query_str += f'{disjunction} prefix == "{addr}" '
            else:
                query_str += \
                    f' {disjunction} ip6AddressList.str.startswith("{addr}/") '
        elif ':' in addr and state.table in ['macs', 'arpnd']:
            query_str += f' {disjunction} macaddr == "{addr}" '
        else:
            try:
                addr = ip_address(addr)
                macaddr = None
            except ValueError:
                macaddr = convert_macaddr_format_to_colon(addr)
                addr = None

            if state.table == "macs":
                query_str = f'{disjunction} macaddr == "{macaddr}" '
            elif state.table == 'arpnd':
                if addr:
                    query_str += f' {disjunction} ipAddress == "{addr}" '
                elif macaddr:
                    query_str += f' {disjunction} macaddr == "{macaddr}" '
            elif state.table == 'routes':
                query_str += f'{disjunction} prefix == "{addr}" '
            else:
                query_str = ''

        if not disjunction:
            disjunction = 'or'

    state.query_str = query_str
    state.unique_query = unique_query
    return query_str, unique_query, columns


def search_sidebar(state, sqobjs):
    '''Draw the sidebar'''

    devdf = gui_get_df(sqobjs['device'], columns=['namespace', 'hostname'])
    if devdf.empty:
        st.error('Unable to retrieve any namespace info')
        st.stop()

    namespaces = [''] + sorted(devdf.namespace.unique().tolist())
    if not state.namespace:
        nsidx = 0
    else:
        nsidx = namespaces.index(state.namespace)
    namespace = st.sidebar.selectbox('Namespace',
                                     namespaces, key='search_ns',
                                     index=nsidx, on_change=search_sync_state)

    st.sidebar.markdown(
        """Displays last 5 search results.

You can use search to find specific objects. You can qualify what you're searching for by qualifying the search term with the type. We support:
- __addresses__: You can qualify a specific table to look for the address. The search string can start with one of the following keywords: __route, mac, arpnd__, to specify which table you want the search to be performed in . If you don't specify a table name, we assume ```network find``` to search for the network attach point for the address. For example, ```arpnd 172.16.1.101``` searches for entries with 172.16.1.101 in the IP address column of the arpnd table. Similarly, ```10.0.0.21``` searches for where in the network that IP address is attached to.
- __ASN__: Start the search with the string ```asn``` followed by the ASN number. Typing ```asns``` will show you the list of unique ASNs across the specified namespaces.
- __VTEP__: Start the search with the string ```vtep``` followed by the VTEP IP address. Typing ```vteps``` will show you the list of unique VTEPs across the specified namespaces.
- __VNI__: Start the search with the string ```vni``` followed by the VNI number. Typing ```vnis``` will show you the list of unique VNIs across the specified namespaces.
- __VLAN__: Start the search with the string ```vlan``` followed by the VLAN number. Typing ```vlans``` will show you the list of unique VLANs across the specified namespaces.
- __MTU__: Start the search with the string ```mtus``` followed by the MTU number. Typing ```mtus``` will show you the list of unique MTUs across the specified namespaces.

When specifying a table, you can specify multiple addresses to look for by providing the addresses as a space separated values such as ```172.16.1.101 10.0.0.11``` or ```mac 00:01:02:03:04:05 00:21:22:23:24:25``` and so on. A combination of mac and IP address can also be specified with the address table. Support for more sophisticated search will be added in the next few releases.
""")

    if namespace != state.namespace:
        state.namespace = namespace

    return namespace


def page_work(state_container):
    '''Main page workhorse'''

    if 'searchSessionState' not in state_container:
        state_container['searchSessionState'] = SearchSessionState()

    state = state_container.searchSessionState

    namespace = search_sidebar(state, state_container.sqobjs)

    query_str, uniq_dict, columns = build_query(state,
                                                state_container.search_text)
    if namespace:
        query_ns = [namespace]
    else:
        query_ns = []
    if query_str:
        if state.table == "network":
            df = gui_get_df(state_container.sqobjs[state.table],
                            verb='find',
                            namespace=query_ns,
                            view="latest", columns=columns,
                            address=query_str)
        else:
            df = gui_get_df(state_container.sqobjs[state.table],
                            namespace=query_ns, query_str=query_str,
                            view="latest", columns=columns)
            if not df.empty:
                df = df.query(query_str) \
                    .drop_duplicates() \
                    .reset_index(drop=True)

        expander = st.expander(f'Search for {state_container.search_text}',
                               expanded=True)
        with expander:
            if not df.empty:
                st.dataframe(df)
            else:
                st.info('No matching result found')
    elif uniq_dict:
        columns = ['namespace'] + uniq_dict['column']
        df = gui_get_df(state_container.sqobjs[uniq_dict['table']],
                        namespace=query_ns, view='latest', columns=columns)
        if not df.empty:
            df = df.groupby(by=columns).first().reset_index()

        expander = st.expander(f'Search for {state_container.search_text}',
                               expanded=True)
        with expander:
            if not df.empty:
                st.dataframe(df)
            else:
                st.info('No matching result found')

    for count, prev_res in enumerate(reversed(state.prev_results)):
        psrch, prev_df = prev_res
        if psrch == state_container.search_text:
            continue
        expander = st.expander(f'Search for {psrch}', expanded=True)
        with expander:
            if not prev_df.empty:
                st.dataframe(prev_df)
            else:
                st.info('No matching result found')

    if ((query_str or uniq_dict) and
            (state_container.search_text != state.search_text)):
        state.prev_results.append((state_container.search_text, df))
        state.search_text = state_container.search_text

    st.experimental_set_query_params(**asdict(state))


def search_sync_state():
    wsstate = st.session_state
    state = wsstate.searchSessionState

    if state.namespace != wsstate.search_ns:
        state.namespace = wsstate.search_ns



from .engineobj import SqPandasEngine
from dataclasses import dataclass
import os

import pandas as pd
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt

from suzieq.sqobjects import interfaces, lldp, bgp, ospf, basicobj, address, evpnVni, arpnd, device, macs
from suzieq.sqobjects.basicobj import SqObject
from suzieq.exceptions import EmptyDataframeError


# TODO:
#  topology for different VRFs?
#  iBGP vs eBGP?
#  color by device type?
#  physical topology without LLDP -- is this possible?
#  how to draw multiple topologies
#  be able to ask if a node has neighbors by type (physical, overlay, protocol, etc)
#  questions
#    * without knowing hierarchy, labels or tags it's unclear how to group things for good picture
# how could we add state of connection (like Established) per protocol

graph_output_dir = '/tmp/suzieq-graphs'


class TopologyObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'topology'

    def _init_dfs(self, namespaces):
        """Initialize the dataframes used"""

        self._if_df = interfaces.IfObj(
            context=self.ctxt).get(
                namespace=namespaces, state="up",
                columns=['namespace', 'hostname', 'ifname', 'ipAddressList',
                         'ip6AddressList', 'state', 'type', 'master', 'macaddr'])

        if self._if_df.empty:
            raise EmptyDataframeError(f"No interface found for {namespaces}")
        self._if_df['vrf'] = self._if_df.apply(
            lambda x: x['master'] if x['type'] not in ['bridge', 'bond_slave']
            else 'default', axis=1)
        self._if_df['vrf'] = np.where(self._if_df['vrf'] == '', 'default',
                                      self._if_df['vrf'])
        self._if_df = self._if_df.explode('ipAddressList') \
                                 .explode('ip6AddressList') \
                                 .fillna({'ipAddressList': '0.0.0.0/0',
                                          'ip6AddressList': '::0/0'})
        self._if_df['plen'] = self._if_df.ipAddressList.str.split('/').str[1]
        self._if_df['ipAddress'] = self._if_df.ipAddressList.str.split(
            '/').str[0]
        self._if_df = self._if_df.drop(
            columns=['ipAddressList', 'ip6AddressList'])

        self.nses = self._if_df['namespace'].unique()

    def get(self, **kwargs):
        self._namespaces = kwargs.get("namespace", self.ctxt.namespace)
        hostname = kwargs.pop('hostname', [])
        user_query = kwargs.pop('query_str', '')
        polled = kwargs.pop('polled', '')
        via = kwargs.pop('via', [])
        ifname = kwargs.pop('ifname', '')
        peerHostname = kwargs.pop('peerHostname', [])
        columns = kwargs.pop('columns', ['default'])

        self._init_dfs(self._namespaces)
        self.lsdb = pd.DataFrame()
        self._a_df = pd.DataFrame()
        self._ip_table = pd.DataFrame()

        self.services = [
            Services('lldp', lldp.LldpObj, {}, ['ifname', 'vrf'],
                     self._augment_lldp_show),
            Services('arpnd', arpnd.ArpndObj, {},
                     ['ipAddress', 'macaddr', 'ifname', 'vrf', 'arpndBidir'],
                     self._augment_arpnd_show),
            Services('bgp', bgp.BgpObj, {'state': 'Established',
                                         'columns': ['*']},
                     ['vrf'],
                     self._augment_bgp_show),
            # Services('evpnVni', evpnVni.EvpnvniObj, {}, 'peerHostname',
            #    'vni', self._augment_evpnvni_show),
            Services('ospf', ospf.OspfObj, {'state': 'full'},
                     ['ifname', 'vrf'], self._augment_ospf_show),
        ]

        if via:
            self.services = [x for x in self.services if x.name in via]
        key = 'peerHostname'
        for srv in self.services:
            if 'columns' not in srv.extra_args:
                srv.extra_args['columns'] = columns
            df = srv.data(context=self.ctxt).get(
                **kwargs,
                **srv.extra_args
            )

            if not df.empty:
                if srv.augment:
                    df = srv.augment(df)
                cols = ['namespace', 'hostname', key]
                if srv.extra_cols:
                    cols = cols + srv.extra_cols
                df = df[cols]
                df.insert(len(df.columns), srv.name, True)
                if self.lsdb.empty:
                    self.lsdb = df
                else:
                    self.lsdb = self.lsdb.merge(df, how='outer')

        self._find_polled_neighbors(polled)
        if self.lsdb.empty:
            return self.lsdb

        if 'ipAddress' not in self.lsdb.columns:
            self.lsdb['ipAddress'] = ''

        self.lsdb = self.lsdb[~self.lsdb.hostname.isna()] \
                        .drop_duplicates() \
                        .rename({'ipAddress': 'peerIP', 'macaddr': 'peerMac'},
                                axis=1, errors='ignore') \
                        .dropna(subset=['peerHostname', 'peerIP'], how='all') \
                        .fillna({'peerHostname': 'unknown',
                                 'ifname': 'unknown', 'arpnd': False,
                                 'peerIP': '', 'peerMac': '',
                                 'arpndBidir': False,
                                 'bgp': False, 'ospf': False,
                                 'lldp': False, 'vrf': 'N/A'})

        self.lsdb['vrf'] = np.where(self.lsdb.vrf == "bridge", "-",
                                    self.lsdb.vrf)
        self.lsdb['peerHostname'] = np.where(
            self.lsdb.peerHostname == "unknown", self.lsdb.peerIP,
            self.lsdb.peerHostname)
        self.lsdb = self.lsdb.drop(columns=['peerIP', 'peerMac'],
                                   errors='ignore')
        self.lsdb = self.lsdb.query('peerHostname != ""')

        # Apply the appropriate filters
        if not self.lsdb.empty:
            if hostname:
                self.lsdb = self.lsdb.query(f'hostname.isin({hostname})')
            if peerHostname:
                self.lsdb = self.lsdb.query(
                    f'peerHostname.isin({peerHostname})')
            if ifname:
                self.lsdb = self.lsdb.query(f'ifname.isin({ifname})')

        if user_query and not self.lsdb.empty:
            self.lsdb = self.lsdb.query(user_query)

        cols = self.lsdb.columns.tolist()
        if 'timestamp' in self.lsdb.columns:
            cols.remove('timestamp')
            cols.append('timestamp')

        return self.lsdb[cols].reset_index(drop=True)

    def _find_polled_neighbors(self, polled):
        if self.lsdb.empty:
            return

        devices = device.DeviceObj(context=self.ctxt).get(namespace=self._namespaces,
                                                          columns=['namespace', 'hostname'])
        self.lsdb = devices.merge(self.lsdb, how='outer',
                                  indicator=True)
        self.lsdb = self.lsdb.rename(columns={'_merge': 'polled'})
        self.lsdb.polled = self.lsdb.polled == 'both'

        if polled and polled.lower() == 'true':
            self.lsdb = self.lsdb.query('polled == True')
        elif polled.lower() == 'false':
            self.lsdb = self.lsdb.query('polled == False')
        if self.lsdb.empty:
            self.lsdb = pd.DataFrame()

    def _create_graphs_from_lsdb(self):
        self.graphs = {}
        for ns, df in self.lsdb.groupby(by=['namespace']):
            attrs = [srv.name for srv in self.services if srv.name in df.columns]
            self.graphs[ns] = nx.from_pandas_edgelist(
                df, 'hostname', 'peerHostname', attrs, nx.MultiGraph)

    # TODO: eventually this needs to move to ospf after we figure out the
    #   schema augmentation story
    def _augment_ospf_show(self, df):
        if not df.empty:
            df = df.query('adjState != "passive"').reset_index(drop=True)

        return df

    def _augment_lldp_show(self, df):
        if not df.empty:
            df = df[df.peerHostname != '']
            df = df.merge(
                self._if_df[['namespace', 'hostname', 'ifname', 'vrf']],
                on=['namespace', 'hostname', 'ifname'])
        return df

    def _augment_bgp_show(self, df):
        return df

    def _augment_evpnvni_show(self, df):
        if not df.empty:
            df = df.explode('remoteVtepList').dropna(how='any')
            df = df.rename(columns={'remoteVtepList': 'peerIP'})
            df = df.merge(self.ip_table, on=['namespace', 'peerIP'],
                          how='left').dropna(how='any').drop_duplicates()
        return df

    def _augment_arpnd_show(self, df):

        if not df.empty:
            # weed out entries that are not reachable
            df = df.query('state != "failed"').reset_index(drop=True)
            # Get the VRF
            df = df.merge(
                self._if_df[['namespace', 'hostname', 'ifname', 'master']],
                left_on=['namespace', 'hostname', 'oif'],
                right_on=['namespace', 'hostname', 'ifname'],
                suffixes=['', '_y'], how='outer') \
                .rename(columns={'master': 'vrf'}) \
                .dropna(subset=['ipAddress'])

            df['vrf'] = np.where(df['vrf'] == '', 'default', df['vrf'])
            df = df.drop(columns=['vlan', 'oif', 'mackey', 'remoteVtepIp',
                                  'timestamp_y'], errors='ignore')

            # address are where I find the neighbor, these own the mac
            addr = self.address_df[['namespace', 'hostname', 'macaddr']] \
                .drop_duplicates() \
                .rename(columns={'hostname': 'peerHostname'})
            df = df.merge(addr, on=['namespace', 'macaddr'], how='left') \
                   .dropna(subset=['hostname', 'ipAddress']) \
                   .rename(columns={'oif': 'ifname'}) \
                   .drop_duplicates()

            # Use MAC table entries to find the local port for a MAC on an SVI
            mac_df = macs.MacsObj(context=self.ctxt) \
                         .get(namespace=self._namespaces, localOnly=True,
                              columns=['namespace', 'hostname', 'vlan',
                                       'macaddr', 'oif'])
            df = df.merge(mac_df,
                          on=['namespace', 'hostname', 'macaddr'],
                          how='outer') \
                .dropna(subset=['hostname'])
            df['ifname'] = np.where(
                df['oif'].isnull(), df['ifname'], df['oif'])

            df['arpndBidir'] = df.apply(
                lambda x, y: True
                if not y.query(f'namespace=="{x.namespace}" and '
                               f'hostname=="{x.peerHostname}" and '
                               f'peerHostname=="{x.hostname}"').empty
                else False, args=(df,), axis=1)

        self._arpnd_df = df
        return self._arpnd_df

    @property
    def address_df(self):
        if self._a_df.empty:
            self._a_df = self._if_df
        return self._a_df

    @property
    def ip_table(self):
        if self._ip_table.empty:
            addr = self.address_df
            if not addr.empty:
                self._ip_table = addr[['namespace',
                                       'hostname', 'ipAddressList']]
                self._ip_table = self._ip_table.explode(
                    'ipAddressList').dropna(how='any')
                self._ip_table = self._ip_table \
                                     .rename(
                                         columns={'ipAddressList': 'peerIP',
                                                  'hostname': 'peerHostname'})
                self._ip_table['peerIP'] = self._ip_table['peerIP'] \
                                               .str.replace("/.+", "")
                self._ip_table = self._ip_table[self._ip_table['peerIP'] != '-']
        return self._ip_table

    def summarize(self, **kwargs):
        self.get(**kwargs)
        if self.lsdb.empty:
            return self.lsdb

        self.ns = {}
        for i in self.nses:
            self.ns[i] = {}
        self._create_graphs_from_lsdb()
        self._analyze_lsdb_graph()

        return pd.DataFrame(self.ns)

    def _analyze_lsdb_graph(self):
        for ns in self.nses:
            for name in [srv.name for srv in self.services if
                         nx.get_edge_attributes(self.graphs[ns], srv.name)]:

                G = nx.Graph([(s, d, data) for s, d, data in
                              self.graphs[ns].edges(data=True) if data[name] == True])
                if G.nodes:
                    self.ns[ns][f'{name}_number_of_nodes'] = len(G.nodes)
                    self.ns[ns][f'{name}_number_of_edges'] = len(G.edges)
                    if not nx.is_connected(G):
                        self.ns[ns][f'{name}_is_fully_connected'] = False
                        self.ns[ns][f'{name}_center'] = False
                    else:
                        self.ns[ns][f'{name}_is_fully_connected'] = True
                        self.ns[ns][f'{name}_center'] = nx.barycenter(G)

                    self.ns[ns][f'{name}_self_loops'] = list(
                        nx.nodes_with_selfloops(G))

                    self.ns[ns][f'{name}_number_of_disjoint_sets'] = len(
                        list(nx.connected_components(G)))

                    self.ns[ns][f'{name}_degree_histogram'] = nx.degree_histogram(
                        G)

                    # if there are too many degrees than the column gets too big
                    if len(self.ns[ns][f'{name}_degree_histogram']) > 6:
                        self.ns[ns][f'{name}_degree_histogram'] = '...'

                else:
                    for k in [f'{name}_is_fully_connected', f'{name}_center',
                              f'{name}_self_loops', f'{name}_number_of_disjoint_sets',
                              f'{name}_degree_histogram', f'{name}_number_of_nodes',
                              f'{name}_number_of_edges']:
                        self.ns[ns][k] = None

    def _make_images(self):
        if not os.path.exists(graph_output_dir):
            os.makedirs(graph_output_dir)
        for ns in self.nses:
            pos = nx.spring_layout(self.graphs[ns])
            for name in [srv.name for srv in self.services if
                         nx.get_edge_attributes(self.graphs[ns], srv.name)]:
                edges = [(u, v) for u, v, d in self.graphs[ns].edges(data=True) if
                         d[name] == True]
                if len(edges) > 1:

                    nx.draw_networkx_nodes(
                        self.graphs[ns], pos=pos, node_size=25)
                    if len(self.graphs[ns].nodes) < 20:
                        nx.draw_networkx_labels(
                            self.graphs[ns], pos=pos, font_size=8)

                    nx.draw_networkx_edges(
                        self.graphs[ns], edgelist=edges, pos=pos)
                    plt.savefig(f"{graph_output_dir}/{ns}_{name}.png")
                    print(f"created {graph_output_dir}/{ns}_{name}.png")
                    plt.close()

    def unique(self, **kwargs):
        '''Unique values for topology'''
        columns = kwargs.pop('columns', ['default'])
        count = kwargs.pop("count", 0)
        query_str = kwargs.get('query_str', '')

        df = self.get(columns=['*'], **kwargs)
        column = columns[0]

        if not count:
            return (pd.DataFrame({f'{column}': df[column].unique()}))
        else:
            r = df[column].value_counts()
            return (pd.DataFrame({column: r})
                    .reset_index()
                    .rename(columns={column: 'numRows',
                                     'index': column})
                    .sort_values(column))


@dataclass(frozen=True)
class Services:
    name: str
    data: SqObject
    extra_args: dict
    extra_cols: list
    augment: any



from suzieq.poller.services.service import Service
import re
from suzieq.utils import convert_macaddr_format_to_colon
from suzieq.utils import expand_nxos_ifname, expand_eos_ifname, expand_ios_ifname
import numpy as np


class MacsService(Service):
    """Mac address service. Different class because of macaddr not in key"""

    def get_key_flds(self):
        """The MAC table in Linux can have weird keys"""
        return ['vlan', 'macaddr', 'oif', 'remoteVtepIp', 'bd']

    def _add_mackey_protocol(self, entry):
        """Construct a key field that is unique.

        This is because of the following cases:
        1. Cumulus has all 0 MAC address with different remoteVtepIP, but
           same VLAM, as ingress replication entries.
        2. Cumulus has interface MAC entries with the same MAC and VLAN 0
        3. Juniper's VPLS entries have only BD, no VLAN
        4. The remaining regular entries where VLAN disambiguates a MAC.
        """
        # Make VLAN int first
        if entry.get('vlan', ''):
            entry['vlan'] = int(entry['vlan'])

        if entry.get('bd', ""):
            if not entry.get('vlan', 0):
                entry['mackey'] = entry['bd']
            else:
                entry['mackey'] = f'{entry["bd"]}-{entry["vlan"]}'
        else:
            if entry.get("remoteVtepIp", ''):
                if entry['macaddr'] == '00:00:00:00:00:00':
                    entry['mackey'] = f'{entry["oif"]}-{entry["remoteVtepIp"]}'
                else:
                    entry['mackey'] = entry['vlan']
            else:
                if entry.get('vlan', 0):
                    entry['mackey'] = entry['vlan']
                else:
                    if entry['flags'] in ['static', 'permanent']:
                        entry['mackey'] = f'{entry["oif"]}'
                    else:
                        entry['mackey'] = '0'

    def _clean_linux_data(self, processed_data, raw_data):
        drop_indices = []
        macentries = {}
        for i, entry in enumerate(processed_data):
            macaddr = entry.get('macaddr', None)
            oif = entry.get('oif', '')
            if macaddr and (macaddr != "00:00:00:00:00:00"):
                key = f'{macaddr}-{oif}'
                old_entry = macentries.get(key, None)
                if not old_entry:
                    macentries[key] = entry
                elif not (old_entry['vlan'] and entry['vlan']):
                    # Ensure we don't munge entries with the diff valid VLANs
                    if not old_entry.get('vlan', ''):
                        old_entry['vlan'] = entry['vlan']
                    else:
                        old_entry['remoteVtepIp'] = entry['remoteVtepIp']
                    old_entry['flags'] = 'remote'
                    self._add_mackey_protocol(old_entry)
                    drop_indices.append(i)
                    continue
            if entry['flags'] == 'offload' or entry['flags'] == 'extern_learn':
                if entry['remoteVtepIp']:
                    entry['flags'] = 'remote'
            self._add_mackey_protocol(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_cumulus_data(self, processed_data, raw_data):
        return self._clean_linux_data(processed_data, raw_data)

    def _clean_junos_data(self, processed_data, raw_data):
        drop_indices = []
        new_entries = []

        for i, entry in enumerate(processed_data):
            if not entry.get('macaddr', ''):
                drop_indices.append(i)
            inflags = entry.pop('flags', '')
            flags = ''
            if inflags:
                if 'rcvd_from_remote' in inflags:
                    flags += ' remote'

            maclist = entry.get('macaddr', '')
            if isinstance(maclist, list):
                # MX format
                oifs = entry.get('oif', [])
                for j, mac in enumerate(maclist):
                    new_entry = entry.copy()
                    new_entry['macaddr'] = mac
                    new_entry['oif'] = oifs[j].split('.')[0]
                    new_entry['flags'] = ''
                    self._add_mackey_protocol(new_entry)
                    new_entries.append(new_entry)
                drop_indices.append(i)
            else:
                entry['flags'] = flags.strip()
                self._add_mackey_protocol(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        processed_data.extend(new_entries)
        return processed_data

    def _clean_nxos_data(self, processed_data, raw_data):

        for entry in processed_data:
            entry['macaddr'] = convert_macaddr_format_to_colon(
                entry.get('macaddr', '0000.0000.0000'))
            vtepIP = re.match(r'(\S+)\(([0-9.]+)\)', entry['oif'])
            if vtepIP:
                entry['remoteVtepIp'] = vtepIP.group(2)
                entry['oif'] = vtepIP.group(1)
                entry['flags'] = 'remote'
            else:
                entry['oif'] = expand_nxos_ifname(entry['oif'])
                entry['remoteVtepIp'] = ''
            if entry.get('vlan', '-') == '-':
                entry['vlan'] = 0
            self._add_mackey_protocol(entry)

        return processed_data

    def _clean_iosxe_data(self, processed_data, raw_data):

        for entry in processed_data:
            entry['macaddr'] = convert_macaddr_format_to_colon(
                entry.get('macaddr', '0000.0000.0000'))
            entry['oif'] = expand_ios_ifname(entry['oif'])
            entry['remoteVtepIp'] = ''
            if entry.get('vlan', ' ').strip() == "All":
                entry['vlan'] = 0
            entry['flags'] = entry.get('flags', '').lower()

            self._add_mackey_protocol(entry)

        return processed_data

    def _clean_ios_data(self, processed_data, raw_data):
        return self._clean_iosxe_data(processed_data, raw_data)

    def _clean_eos_data(self, processed_data, raw_data):

        drop_indices = []

        for i, entry in enumerate(processed_data):
            entry['oif'] = expand_eos_ifname(entry['oif'])
            if entry['oif'].startswith('Vxlan') and 'remoteVtepIp' not in entry:
                drop_indices.append(i)
                continue
            if '.' in entry['macaddr']:
                entry['macaddr'] = convert_macaddr_format_to_colon(
                    entry.get('macaddr', '0000.0000.0000'))
            self._add_mackey_protocol(entry)

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data



from .engineobj import SqPandasEngine


class DevconfigObj(SqPandasEngine):

    def __init__(self, baseobj):
        super().__init__(baseobj)

    @staticmethod
    def table_name():
        return 'devconfig'



import logging
from abc import ABC, abstractmethod
from typing import List
from dataclasses import dataclass

import pandas as pd


class SqDB(ABC):
    def __init__(self, cfg: dict, logger: logging.Logger):
        """Initialize the database

        :param cfg: dict, Suzieq config dictionary
        :param logger: logging.Logger, logger to use for logging
        :returns: Self
        :rtype: SqDB

        """
        self.cfg = cfg
        self.logger = logger or logging.getLogger()

    @abstractmethod
    def get_tables(self):
        """Get a list of tables in the database

        Raises:
            NotImplementedError: This method is not implemented
        """
        raise NotImplementedError

    @abstractmethod
    def read(self, table_name: str, data_format: str,
             **kwargs) -> pd.DataFrame:
        """This is the routine that reads the DB and returns a pandas DF

        :param table_name: str, Name of the table to read data for
        :param data_format: str, Format the data's to be returned in,
                            (only pandas supported at this point)
        :returns: data in the form of a pandas dataframe
        :rtype: pd.DataFrame
        """
        raise NotImplementedError

    @abstractmethod
    def write(self, table_name: str, data_format: str,
              data, **kwargs) -> int:
        """Write the data supplied as a dataframe in the appropriate format

        :param cfg: Suzieq configuration
        :param table_name: str, Name of the table to write data to
        :param data: data to be written, usually pandas DF, but can be
                     engine specific (spark, dask etc.)
        :param data_format: str, Format the data's to be returned in,
                            (only pandas supported at this point)
        :returns: status of write
        :rtype: integer

        """
        raise NotImplementedError

    @abstractmethod
    def coalesce(self, tables: List[str] = None, period: str = '') -> None:
        """Coalesce the database files in specified folder.

        :param tables: List[str], List of specific tables to coalesce, empty for all
        : param period: str, period value to override what the configuration states
        :returns: Nothing

        """
        raise NotImplementedError


@dataclass
class SqCoalesceStats:
    service: str
    period: str                 # Coalescing period
    execTime: float            # execution time in secs
    fileCount: int          # number of files written
    recCount: int           # number of written records
    timestamp: int              # when this record was created



import typing
import os
from nubia import command, argument, context


@command("set")
@argument("namespace", description="namespace to qualify selection")
@argument("hostname", description="Name of host to qualify selection")
@argument(
    "start_time", description="Start of time window in YYYY-MM-dd HH:mm:SS format"
)
@argument("end_time", description="End of time window in YYYY-MM-dd HH:mm:SS format")
@argument("pager", description="Enable pagination prompt on longer outputs",
          choices=['on', 'off'])
@argument('col_width', description='Max Width of each column in table display')
@argument(
    "engine",
    choices=["pandas"],
    description="Use Pandas for non-SQL commands",
)
@argument(
    "datadir",
    description="Set the data directory for the command"
)
def set_ctxt(
        pager: str = 'on',
        hostname: typing.List[str] = [],
        start_time: str = "",
        end_time: str = "",
        namespace: typing.List[str] = [],
        engine: str = "",
        datadir: str = "",
        col_width: int = 50,
):
    """set certain contexts for subsequent commands. Cmd is additive"""
    plugin_ctx = context.get_context()

    if namespace:
        plugin_ctx.namespace = namespace

    if hostname:
        plugin_ctx.hostname = hostname

    if start_time:
        plugin_ctx.start_time = start_time

    if end_time:
        plugin_ctx.end_time = end_time

    if engine:
        plugin_ctx.change_engine(engine)

    if datadir:
        if os.path.isdir(datadir):
            plugin_ctx.cfg['data-directory'] = datadir
        else:
            print(f'{datadir} is not a valid directory')

    if col_width:
        plugin_ctx.col_width = int(col_width)

    if pager == 'on':
        plugin_ctx.pager = True


@command("clear")
@argument("namespace", description="namespace to qualify selection")
@argument("hostname", description="Name of host to qualify selection")
@argument(
    "start_time", description="Start of time window in YYYY-MM-dd HH:mm:SS format"
)
@argument("end_time", description="End of time window in YYYY-MM-dd HH:mm:SS format")
@argument("pager", description="End of time window in YYYY-MM-dd HH:mm:SS format")
def clear_ctxt(
        pager: str = 'off',
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        namespace: str = "",
):
    """clear certain contexts for subsequent commands. Cmd is additive"""
    plugin_ctx = context.get_context()

    if namespace:
        plugin_ctx.namespace = []

    if hostname:
        plugin_ctx.hostname = []

    if start_time:
        plugin_ctx.start_time = ""

    if end_time:
        plugin_ctx.end_time = ""

    if pager:
        plugin_ctx.pager = False



import time
import re
from nubia import command, argument

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.inventory import InventoryObj


@command("inventory", help="Act on LLDP data")
class InventoryCmd(SqCommand):
    """Device inventory information such as serial number, cable info etc"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        query_str: str = ' ',
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=InventoryObj,
        )

    @command("show")
    @argument("type", description="Filter by type",
              choices=["fan", "power", "xcvr", "supervisor", "port-adapter", "linecard", "fabric", "midplane", "mx-cb"])
    @argument("status", description="Filter by status",
              choices=['present', 'absent'])
    @argument("model", description="Filter by model")
    @argument("serial", description="Filter by serial number")
    @argument("vendor", description="Filter by vendor name")
    def show(self, type: str = "", status: str = "", model: str = "",
             serial: str = "", vendor: str = "") -> None:
        """Show Device inventory info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        if vendor:
            vendor = re.split(r"\s+(?=[^']*(?:'))", vendor)
        else:
            vendor = []

        if model:
            model = re.split(r"\s+(?=[^']*(?:'))", model)
        else:
            model = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                columns=self.columns,
                                query_str=self.query_str,
                                type=type.split(),
                                status=status,
                                model=model,
                                vendor=vendor,
                                serial=serial.split(),
                                namespace=self.namespace,
                                )
        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)



from datetime import timedelta, datetime, timezone

from suzieq.poller.services.service import Service
from suzieq.utils import get_timestamp_from_junos_time
from dateparser import parse
from datetime import datetime


class DeviceService(Service):
    """Checks the uptime and OS/version of the node.
    This is specially called out to normalize the timestamp and handle
    timestamp diff
    """

    def __init__(self, name, defn, period, stype, keys, ignore_fields,
                 schema, queue, run_once):
        super().__init__(name, defn, period, stype, keys, ignore_fields,
                         schema, queue, run_once)
        self.ignore_fields.append("bootupTimestamp")

    def _common_data_cleaner(self, processed_data, raw_data):
        for entry in processed_data:
            entry['status'] = "alive"
            entry["address"] = raw_data[0]["address"]

        return processed_data

    def _clean_linux_data(self, processed_data, raw_data):

        for entry in processed_data:
            # We're assuming that if the entry doesn't provide the
            # bootupTimestamp field but provides the sysUptime field,
            # we fix the data so that it is always bootupTimestamp
            # TODO: Fix the clock drift
            if not entry.get("bootupTimestamp", None) and entry.get(
                    "sysUptime", None):
                entry["bootupTimestamp"] = int(
                    int(raw_data[0]["timestamp"])/1000 -
                    float(entry.pop("sysUptime", 0))
                )
                if entry["bootupTimestamp"] < 0:
                    entry["bootupTimestamp"] = 0
            # This is the case for Linux servers, so also extract the vendor
            # and version from the os string
            if not entry.get("vendor", ''):
                if 'os' in entry:
                    osstr = entry.get("os", "").split()
                    if len(osstr) > 1:
                        # Assumed format is: Ubuntu 18.04.2 LTS,
                        # CentOS Linux 7 (Core)
                        entry["vendor"] = osstr[0]
                        if not entry.get("version", ""):
                            entry["version"] = ' '.join(osstr[1:])
                    del entry["os"]
            entry['os'] = 'linux'

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_cumulus_data(self, processed_data, raw_data):
        for entry in processed_data:
            model = entry.get('_modelName', '')
            if model:
                entry['model'] = model
            uptime = entry.get('_uptime', '').split()
            if uptime:
                hr, mins, secs = uptime[-1].split(':')
                if len(uptime) > 1:
                    days = int(uptime[0])
                else:
                    days = 0
                uptime_delta = timedelta(days=days, hours=int(hr),
                                         minutes=int(mins),
                                         seconds=int(secs.split('.')[0]))
                entry['bootupTimestamp'] = int(
                    (int(raw_data[0]["timestamp"])/1000) -
                    uptime_delta.total_seconds())

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_sonic_data(self, processed_data, raw_data):
        return self._clean_linux_data(processed_data, raw_data)

    def _clean_common_ios(self, entry, os):
        '''Common IOS-like NOS cleaning'''
        entry['os'] = os
        entry['vendor'] = 'Cisco'
        if entry.get('bootupTimestamp', ''):
            entry['bootupTimestamp'] = int(datetime.utcfromtimestamp(
                parse(entry['bootupTimestamp']).timestamp()).timestamp())

    def _clean_iosxr_data(self, processed_data, raw_data):
        for entry in processed_data:
            self._clean_common_ios(entry, 'iosxr')
            if 'IOS-XRv' in entry.get('model', ''):
                entry['architecture'] = "x86-64"

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_iosxe_data(self, processed_data, raw_data):
        for entry in processed_data:
            self._clean_common_ios(entry, 'iosxe')

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_ios_data(self, processed_data, raw_data):
        for entry in processed_data:
            self._clean_common_ios(entry, 'ios')

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_junos_data(self, processed_data, raw_data):

        for entry in processed_data:
            entry['bootupTimestamp'] = get_timestamp_from_junos_time(
                entry['bootupTimestamp'], int(raw_data[0]["timestamp"])/1000)/1000

        return self._common_data_cleaner(processed_data, raw_data)

    def _clean_nxos_data(self, processed_data, raw_data):
        for entry in processed_data:
            upsecs = (24*3600*int(entry.pop('kern_uptm_days', 0)) +
                      3600*int(entry.pop('kern_uptm_hrs', 0)) +
                      60*int(entry.pop('kern_uptm_mins', 0)) +
                      int(entry.pop('kern_uptm_secs', 0)))
            if upsecs:
                entry['bootupTimestamp'] = int(
                    int(raw_data[0]["timestamp"])/1000 - upsecs)

        return self._common_data_cleaner(processed_data, raw_data)



import re

from suzieq.poller.services.service import Service


class ConfigService(Service):
    """config service. Different class because minor munging of output"""

    def _common_data_cleaner(self, processed_data, raw_data):
        # The data we get is just a long string that is the config
        # No processing has been performed on the raw data. So processed_data
        # is empty

        output = raw_data[0].get('data', {})
        if isinstance(output, str):
            return []

        if 'output' in output.keys():
            data = output.get('output', "")
        else:
            data = raw_data[0].get('data', '')

        lines = data.splitlines()
        lines = list(filter(
            lambda x: not re.search(
                'Time|username|crypto|key-hash|encryption', x),
            lines))

        processed_data = [{'config': '\n'.join(lines)}]

        return processed_data

    def get_diff(self, old, new):
        """Compare string hashes for fast matches of config
        """

        adds = []
        dels = []

        if old:
            oldcfg = old[0].get('config', '')
        else:
            oldcfg = ''

        if new:
            newcfg = new[0].get('config', '')
        else:
            newcfg = ''

        if oldcfg and not newcfg:
            dels = new
        elif newcfg and not oldcfg:
            adds = new
        else:
            hash_old = hash(oldcfg)
            hash_new = hash(newcfg)
            if hash_old != hash_new:
                adds = new

        return adds, dels
